"""Implements the :stac-ext:`Scientific Citation Extension <scientific>`.

For a description of Digital Object Identifiers (DOIs), see the DOI Handbook:

https://doi.org/10.1000/182
"""

from __future__ import annotations

from typing import Any, Generic, Literal, TypeVar, cast

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.utils import StringEnum, map_opt

#: Generalized version of :class:`~pystac.Collection` or :class:`~pystac.Item`
T = TypeVar("T", pystac.Collection, pystac.Item)

SCHEMA_URI: str = "https://stac-extensions.github.io/scientific/v1.0.0/schema.json"
PREFIX: str = "sci:"

# Field names
DOI_PROP: str = PREFIX + "doi"
CITATION_PROP: str = PREFIX + "citation"
PUBLICATIONS_PROP: str = PREFIX + "publications"

DOI_URL_BASE = "https://doi.org/"


# Link rel type.
class ScientificRelType(StringEnum):
    """A list of rel types defined in the Scientific Citation Extension.

    See the :stac-ext:`Scientific Citation Extension Relation types
    <scientific#relation-types>` documentation for details.
    """

    CITE_AS = "cite-as"
    """Used to indicate a link to the publication referenced by the ``sci:doi``
    field."""


def doi_to_url(doi: str) -> str:
    """Converts a DOI to the corresponding URL."""
    from urllib import parse

    return DOI_URL_BASE + parse.quote(doi)


class Publication:
    """Helper for Publication entries."""

    citation: str | None
    doi: str | None

    def __init__(self, doi: str | None, citation: str | None) -> None:
        self.doi = doi
        self.citation = citation

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Publication):
            return NotImplemented

        return self.doi == other.doi and self.citation == other.citation

    def __repr__(self) -> str:
        return f"<Publication doi={self.doi} target={self.citation}>"

    def to_dict(self) -> dict[str, str | None]:
        import copy

        return copy.deepcopy({"doi": self.doi, "citation": self.citation})

    @staticmethod
    def from_dict(d: dict[str, str]) -> Publication:
        return Publication(d.get("doi"), d.get("citation"))

    def get_link(self) -> pystac.Link | None:
        """Gets a :class:`~pystac.Link` for the DOI for this publication. If
        :attr:`Publication.doi` is ``None``, this method will also return ``None``."""
        if self.doi is None:
            return None
        return pystac.Link(ScientificRelType.CITE_AS, doi_to_url(self.doi))


def remove_link(links: list[pystac.Link], doi: str | None) -> None:
    if doi is None:
        return
    url = doi_to_url(doi)
    for i, a_link in enumerate(links):
        if a_link.rel != ScientificRelType.CITE_AS:
            continue
        if a_link.target == url:
            del links[i]
            break


class ScientificExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` or a :class:`pystac.Collection` with properties from the
    :stac-ext:`Scientific Citation Extension <scientific>`. This class is generic over
    the type of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Collection`).

    To create a concrete instance of :class:`ScientificExtension`, use the
    :meth:`ScientificExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> sci_ext = ScientificExtension.ext(item)
    """

    name: Literal["sci"] = "sci"
    obj: pystac.STACObject

    def __init__(self, obj: pystac.STACObject) -> None:
        self.obj = obj

    def apply(
        self,
        doi: str | None = None,
        citation: str | None = None,
        publications: list[Publication] | None = None,
    ) -> None:
        """Applies scientific extension properties to the extended
        :class:`~pystac.Item`.

        Args:
            doi : Optional DOI string for the item.  Must not be a DOI link.
            citation : Optional human-readable reference.
            publications : Optional list of relevant publications
                referencing and describing the data.
        """
        self.doi = doi
        self.citation = citation
        self.publications = publications

    @property
    def doi(self) -> str | None:
        """Get or sets the DOI for the item.

        This MUST NOT be a DOIs link. For all DOI names respective DOI links SHOULD be
        added to the links section.
        """
        return self._get_property(DOI_PROP, str)

    @doi.setter
    def doi(self, v: str | None) -> None:
        if DOI_PROP in self.properties:
            if v == self.properties[DOI_PROP]:
                return
            remove_link(self.obj.links, self.properties[DOI_PROP])

        if v is not None:
            self.properties[DOI_PROP] = v
            url = doi_to_url(v)
            self.obj.add_link(pystac.Link(ScientificRelType.CITE_AS, url))

    @property
    def citation(self) -> str | None:
        """Get or sets the recommended human-readable reference (citation) to be used by
        publications citing the data.

        No specific citation style is suggested, but the citation should contain all
        information required to find the publication distinctively.
        """
        return self._get_property(CITATION_PROP, str)

    @citation.setter
    def citation(self, v: str | None) -> None:
        self._set_property(CITATION_PROP, v)

    @property
    def publications(self) -> list[Publication] | None:
        """Get or sets the list of relevant publications referencing and describing the
        data."""
        return map_opt(
            lambda pubs: [Publication.from_dict(pub) for pub in pubs],
            self._get_property(PUBLICATIONS_PROP, list[dict[str, Any]]),
        )

    @publications.setter
    def publications(self, v: list[Publication] | None) -> None:
        self._set_property(
            PUBLICATIONS_PROP, map_opt(lambda pubs: [pub.to_dict() for pub in pubs], v)
        )
        if v is not None:
            for pub in v:
                pub_link = pub.get_link()
                if pub_link is not None:
                    self.obj.add_link(pub_link)

    # None for publication will clear all.
    def remove_publication(self, publication: Publication | None = None) -> None:
        """Removes the given :class:`Publication` from the extended
        :class:`~pystac.Item`. If the ``publication`` argument is ``None``, all
        publications will be removed from the :class:`~pystac.Item`."""
        if PUBLICATIONS_PROP not in self.properties:
            return

        if not publication:
            pubs = self.publications
            if pubs is not None:
                for one_pub in pubs:
                    remove_link(self.obj.links, one_pub.doi)

            del self.properties[PUBLICATIONS_PROP]
            return

        # One publication and link to remove
        remove_link(self.obj.links, publication.doi)
        to_remove = publication.to_dict()
        self.properties[PUBLICATIONS_PROP].remove(to_remove)

        if not self.properties[PUBLICATIONS_PROP]:
            del self.properties[PUBLICATIONS_PROP]

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> ScientificExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Scientific
        Extension <scientific>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Collection`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(ScientificExtension[T], CollectionScientificExtension(obj))
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(ScientificExtension[T], ItemScientificExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesScientificExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesScientificExtension(obj)


class CollectionScientificExtension(ScientificExtension[pystac.Collection]):
    """A concrete implementation of :class:`ScientificExtension` on an
    :class:`~pystac.Collection` that extends the properties of the Item to include
    properties defined in the :stac-ext:`Scientific Citation Extension <scientific>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`ScientificExtension.ext` on an :class:`~pystac.Collection` to extend it.
    """

    collection: pystac.Collection
    """The :class:`~pystac.Collection` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Collection` properties, including extension properties."""

    links: list[pystac.Link]
    """The list of :class:`~pystac.Link` objects associated with the
    :class:`~pystac.Collection` being extended, including links added by this extension.
    """

    def __init__(self, collection: pystac.Collection):
        self.collection = collection
        self.properties = collection.extra_fields
        self.links = collection.links
        super().__init__(self.collection)

    def __repr__(self) -> str:
        return "<CollectionScientificExtension Collection id={}>".format(
            self.collection.id
        )


class ItemScientificExtension(ScientificExtension[pystac.Item]):
    """A concrete implementation of :class:`ScientificExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include
    properties defined in the :stac-ext:`Scientific Citation Extension
    <scientific>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`ScientificExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    links: list[pystac.Link]
    """The list of :class:`~pystac.Link` objects associated with the
    :class:`~pystac.Item` being extended, including links added by this extension.
    """

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties
        self.links = item.links
        super().__init__(self.item)

    def __repr__(self) -> str:
        return f"<ItemScientificExtension Item id={self.item.id}>"


class SummariesScientificExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Scientific Citation Extension <scientific>`.
    """

    @property
    def citation(self) -> list[str] | None:
        """Get or sets the summary of :attr:`ScientificExtension.citation` values
        for this Collection.
        """
        return self.summaries.get_list(CITATION_PROP)

    @citation.setter
    def citation(self, v: list[str] | None) -> None:
        self._set_summary(CITATION_PROP, v)

    @property
    def doi(self) -> list[str] | None:
        """Get or sets the summary of :attr:`ScientificExtension.citation` values
        for this Collection.
        """
        return self.summaries.get_list(DOI_PROP)

    @doi.setter
    def doi(self, v: list[str] | None) -> None:
        self._set_summary(DOI_PROP, v)


class ScientificExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {"scientific"}
    stac_object_types = {pystac.STACObjectType.COLLECTION, pystac.STACObjectType.ITEM}


SCIENTIFIC_EXTENSION_HOOKS: ExtensionHooks = ScientificExtensionHooks()



from functools import lru_cache


@lru_cache
def get_jinja_env():  # type: ignore
    try:
        from jinja2 import Environment, PackageLoader, select_autoescape
    except ModuleNotFoundError:
        return None

    environment = Environment(
        loader=PackageLoader("pystac", "html"), autoescape=select_autoescape()
    )

    return environment



import json
import os
import shutil
import tempfile

from pystac import Item, StacIO

from ._base import Bench
from ._util import get_data_path


class ItemBench(Bench):
    def setup(self) -> None:
        self.temp_dir = tempfile.mkdtemp()

        self.stac_io = StacIO.default()

        # using an item with many assets to better test deserialization timing
        self.item_path = get_data_path("eo/eo-sentinel2-item.json")
        with open(self.item_path) as src:
            self.item_dict = json.load(src)
        self.item = Item.from_file(self.item_path)

    def teardown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def time_item_from_file(self) -> None:
        """Deserialize an Item from file"""
        _ = Item.from_file(self.item_path)

    def time_item_from_dict(self) -> None:
        """Deserialize an Item from dictionary."""
        _ = Item.from_dict(self.item_dict)

    def time_item_to_dict(self) -> None:
        """Serialize an Item to a dictionary."""
        self.item.to_dict(include_self_link=True)

    def time_item_save(self) -> None:
        """Serialize an Item to a JSON file."""
        self.item.save_object(
            include_self_link=True,
            dest_href=os.path.join(self.temp_dir, "time_item_save.json"),
            stac_io=self.stac_io,
        )



from pystac.utils import StringEnum


class RelType(StringEnum):
    """A list of common rel types that can be used in STAC Link metadata.
    See :stac-spec:`"Using Relation Types <best-practices.md#using-relation-types>`
    in the STAC Best Practices for guidelines on using relation types. You may also want
    to refer to the "Relation type" documentation for
    :stac-spec:`Catalogs <catalog-spec/catalog-spec.md#relation-types>`,
    :stac-spec:`Collections <collection-spec/collection-spec.md#relation-types>`,
    or :stac-spec:`Items <item-spec/item-spec.md#relation-types>` for relation types
    specific to those STAC objects.
    """

    ALTERNATE = "alternate"
    CANONICAL = "canonical"
    CHILD = "child"
    COLLECTION = "collection"
    ITEM = "item"
    ITEMS = "items"
    LICENSE = "license"
    DERIVED_FROM = "derived_from"
    NEXT = "next"
    PARENT = "parent"
    PREV = "prev"
    PREVIEW = "preview"
    ROOT = "root"
    SELF = "self"
    VIA = "via"



from __future__ import annotations

import json
import os
import tempfile
from collections.abc import Iterator
from copy import deepcopy
from datetime import datetime
from typing import Any

import pytest
from dateutil import tz

import pystac
from pystac import (
    Asset,
    Catalog,
    CatalogType,
    Collection,
    Extent,
    Item,
    ItemCollection,
    Provider,
    SpatialExtent,
    TemporalExtent,
)
from pystac.extensions.eo import EOExtension
from pystac.utils import datetime_to_str, get_required, str_to_datetime
from pystac.validation import validate_dict
from tests.utils import ARBITRARY_BBOX, ARBITRARY_GEOM, TestCases

TEST_DATETIME = datetime(2020, 3, 14, 16, 32)


def test_provider_to_from_dict() -> None:
    provider_dict = {
        "name": "Remote Data, Inc",
        "description": "Producers of awesome spatiotemporal assets",
        "roles": ["producer", "processor"],
        "url": "http://remotedata.io",
        "extension:field": "some value",
    }
    expected_extra_fields = {"extension:field": provider_dict["extension:field"]}

    provider = Provider.from_dict(provider_dict)

    assert (
        provider_dict["name"],
        provider_dict["description"],
        provider_dict["roles"],
        provider_dict["url"],
        expected_extra_fields,
        provider_dict,
    ) == (
        provider.name,
        provider.description,
        provider.roles,
        provider.url,
        provider.extra_fields,
        provider.to_dict(),
    )


def test_spatial_extent_from_coordinates() -> None:
    extent = SpatialExtent.from_coordinates(ARBITRARY_GEOM["coordinates"])

    assert len(extent.bboxes) == 1
    bbox = extent.bboxes[0]
    assert len(bbox) == 4
    for x in bbox:
        assert isinstance(x, float)


def test_read_eo_items_are_heritable() -> None:
    cat = TestCases.case_5()
    item = next(cat.get_items(recursive=True))

    assert EOExtension.has_extension(item)


def test_save_uses_previous_catalog_type() -> None:
    collection = TestCases.case_8()
    assert collection.STAC_OBJECT_TYPE == pystac.STACObjectType.COLLECTION
    assert collection.catalog_type == CatalogType.SELF_CONTAINED
    with tempfile.TemporaryDirectory() as tmp_dir:
        collection.normalize_hrefs(tmp_dir)
        href = collection.self_href
        collection.save()

        collection2 = pystac.Collection.from_file(href)
        assert collection2.catalog_type == CatalogType.SELF_CONTAINED


def test_clone_uses_previous_catalog_type() -> None:
    catalog = TestCases.case_8()
    assert catalog.catalog_type == CatalogType.SELF_CONTAINED
    clone = catalog.clone()
    assert clone.catalog_type == CatalogType.SELF_CONTAINED


def test_clone_cant_mutate_original() -> None:
    collection = TestCases.case_8()
    assert collection.keywords == ["disaster", "open"]
    clone = collection.clone()
    clone.extra_fields["test"] = "extra"
    assert "test" not in collection.extra_fields
    assert clone.keywords is not None
    clone.keywords.append("clone")
    assert clone.keywords == ["disaster", "open", "clone"]
    assert collection.keywords == ["disaster", "open"]
    assert id(collection.summaries) != id(clone.summaries)


def test_multiple_extents() -> None:
    cat1 = TestCases.case_1()
    country = cat1.get_child("country-1")
    assert country is not None
    col1 = country.get_child("area-1-1")
    assert col1 is not None
    col1.validate()
    assert isinstance(col1, Collection)
    validate_dict(col1.to_dict(), pystac.STACObjectType.COLLECTION)

    multi_ext_uri = TestCases.get_path("data-files/collections/multi-extent.json")
    with open(multi_ext_uri) as f:
        multi_ext_dict = json.load(f)
    validate_dict(multi_ext_dict, pystac.STACObjectType.COLLECTION)
    assert isinstance(Collection.from_dict(multi_ext_dict), Collection)

    multi_ext_col = Collection.from_file(multi_ext_uri)
    multi_ext_col.validate()
    ext = multi_ext_col.extent
    extent_dict = multi_ext_dict["extent"]
    assert isinstance(ext, Extent)
    assert isinstance(ext.spatial.bboxes[0], list)
    assert len(ext.spatial.bboxes) == 3
    assert ext.to_dict() == extent_dict

    cloned_ext = ext.clone()
    assert cloned_ext.to_dict() == multi_ext_dict["extent"]


def test_extra_fields() -> None:
    catalog = TestCases.case_2()
    collection = catalog.get_child("1a8c1632-fa91-4a62-b33e-3a87c2ebdf16")
    assert collection is not None

    collection.extra_fields["test"] = "extra"

    with tempfile.TemporaryDirectory() as tmp_dir:
        p = os.path.join(tmp_dir, "collection.json")
        collection.save_object(include_self_link=False, dest_href=p)
        with open(p) as f:
            col_json = json.load(f)
        assert "test" in col_json
        assert col_json["test"] == "extra"

        read_col = pystac.Collection.from_file(p)
        assert "test" in read_col.extra_fields
        assert read_col.extra_fields["test"] == "extra"


def test_update_extents() -> None:
    catalog = TestCases.case_2()
    base_collection = catalog.get_child("1a8c1632-fa91-4a62-b33e-3a87c2ebdf16")
    assert isinstance(base_collection, Collection)
    base_extent = base_collection.extent
    collection = base_collection.clone()

    item1 = Item(
        id="test-item-1",
        geometry=ARBITRARY_GEOM,
        bbox=[-180, -90, 180, 90],
        datetime=TEST_DATETIME,
        properties={"key": "one"},
        stac_extensions=["eo", "commons"],
    )

    item2 = Item(
        id="test-item-1",
        geometry=ARBITRARY_GEOM,
        bbox=[-180, -90, 180, 90],
        datetime=None,
        properties={
            "start_datetime": datetime_to_str(datetime(2000, 1, 1, 12, 0, 0, 0)),
            "end_datetime": datetime_to_str(datetime(2000, 2, 1, 12, 0, 0, 0)),
        },
        stac_extensions=["eo", "commons"],
    )

    collection.add_item(item1)

    collection.update_extent_from_items()
    assert [[-180, -90, 180, 90]] == collection.extent.spatial.bboxes
    assert len(base_extent.spatial.bboxes[0]) == len(
        collection.extent.spatial.bboxes[0]
    )
    assert base_extent.temporal.intervals != collection.extent.temporal.intervals

    collection.remove_item("test-item-1")
    collection.update_extent_from_items()
    assert [[-180, -90, 180, 90]] != collection.extent.spatial.bboxes
    collection.add_item(item2)

    collection.update_extent_from_items()

    assert [
        [
            item2.common_metadata.start_datetime,
            base_extent.temporal.intervals[0][1],
        ]
    ] == collection.extent.temporal.intervals


def test_supplying_href_in_init_does_not_fail() -> None:
    test_href = "http://example.com/collection.json"
    spatial_extent = SpatialExtent(bboxes=[ARBITRARY_BBOX])
    temporal_extent = TemporalExtent(intervals=[[TEST_DATETIME, None]])

    collection_extent = Extent(spatial=spatial_extent, temporal=temporal_extent)
    collection = Collection(
        id="test", description="test desc", extent=collection_extent, href=test_href
    )

    assert collection.get_self_href() == test_href


def test_collection_with_href_caches_by_href() -> None:
    collection = pystac.Collection.from_file(
        TestCases.get_path("data-files/examples/hand-0.8.1/collection.json")
    )
    cache = collection._resolved_objects

    # Since all of our STAC objects have HREFs, everything should be
    # cached only by HREF
    assert len(cache.id_keys_to_objects) == 0


@pytest.mark.block_network
def test_assets() -> None:
    path = TestCases.get_path("data-files/collections/with-assets.json")
    with open(path) as f:
        data = json.load(f)
    collection = pystac.Collection.from_dict(data)
    collection.validate()


def test_get_assets() -> None:
    collection = pystac.Collection.from_file(
        TestCases.get_path("data-files/collections/with-assets.json")
    )

    media_type_filter = collection.get_assets(media_type=pystac.MediaType.PNG)
    assert list(media_type_filter.keys()) == ["thumbnail"]
    role_filter = collection.get_assets(role="thumbnail")
    assert list(role_filter.keys()) == ["thumbnail"]
    multi_filter = collection.get_assets(
        media_type=pystac.MediaType.PNG, role="thumbnail"
    )
    assert list(multi_filter.keys()) == ["thumbnail"]

    no_filter = collection.get_assets()
    assert no_filter is not collection.assets
    assert list(no_filter.keys()) == ["thumbnail"]
    no_filter["thumbnail"].description = "foo"
    assert collection.assets["thumbnail"].description != "foo"

    no_assets = collection.get_assets(media_type=pystac.MediaType.HDF)
    assert no_assets == {}


def test_removing_optional_attributes() -> None:
    path = TestCases.get_path("data-files/collections/with-assets.json")
    with open(path) as file:
        data = json.load(file)
    data["title"] = "dummy title"
    data["stac_extensions"] = ["dummy extension"]
    data["keywords"] = ["key", "word"]
    data["providers"] = [{"name": "pystac"}]
    collection = pystac.Collection.from_dict(data)

    # Assert we have everything set
    assert collection.title
    assert collection.stac_extensions
    assert collection.keywords
    assert collection.providers
    assert collection.summaries
    assert collection.assets

    # Remove all of the optional stuff
    collection.title = None
    collection.stac_extensions = []
    collection.keywords = []
    collection.providers = []
    collection.summaries = pystac.Summaries({})
    collection.assets = {}

    collection_as_dict = collection.to_dict()
    for key in (
        "title",
        "stac_extensions",
        "keywords",
        "providers",
        "summaries",
        "assets",
    ):
        assert key not in collection_as_dict


def test_from_dict_preserves_dict() -> None:
    path = TestCases.get_path("data-files/collections/with-assets.json")
    with open(path) as f:
        collection_dict = json.load(f)
    param_dict = deepcopy(collection_dict)

    # test that the parameter is preserved
    _ = Collection.from_dict(param_dict)
    assert param_dict == collection_dict

    # assert that the parameter is not preserved with
    # non-default parameter
    _ = Collection.from_dict(param_dict, preserve_dict=False, migrate=False)
    assert param_dict != collection_dict


def test_from_dict_set_root() -> None:
    path = TestCases.get_path("data-files/examples/hand-0.8.1/collection.json")
    with open(path) as f:
        collection_dict = json.load(f)
    catalog = pystac.Catalog(id="test", description="test desc")
    collection = Collection.from_dict(collection_dict, root=catalog)
    assert collection.get_root() is catalog


def test_schema_summary() -> None:
    collection = pystac.Collection.from_file(
        TestCases.get_path(
            "data-files/examples/1.0.0/collection-only/collection-with-schemas.json"
        )
    )
    instruments_schema = get_required(
        collection.summaries.get_schema("instruments"),
        collection.summaries,
        "instruments",
    )

    assert isinstance(instruments_schema, dict)


def test_from_invalid_dict_raises_exception() -> None:
    stac_io = pystac.StacIO.default()
    catalog_dict = stac_io.read_json(
        TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
    )
    with pytest.raises(pystac.STACTypeError):
        _ = pystac.Collection.from_dict(catalog_dict)


def test_clone_preserves_assets() -> None:
    path = TestCases.get_path("data-files/collections/with-assets.json")
    original_collection = Collection.from_file(path)
    assert len(original_collection.assets) > 0
    assert all(
        asset.owner is original_collection
        for asset in original_collection.assets.values()
    )

    cloned_collection = original_collection.clone()

    for key in original_collection.assets:
        assert key in cloned_collection.assets, f"Failed to Preserve {key} asset"
        cloned_asset = cloned_collection.assets.get(key)
        if cloned_asset is not None:
            assert cloned_asset.owner is cloned_collection, (
                f"Failed to set owner for {key}"
            )


def test_to_dict_no_self_href() -> None:
    temporal_extent = TemporalExtent(intervals=[[TEST_DATETIME, None]])
    spatial_extent = SpatialExtent(bboxes=ARBITRARY_BBOX)
    extent = Extent(spatial=spatial_extent, temporal=temporal_extent)
    collection = Collection(id="an-id", description="A test Collection", extent=extent)
    d = collection.to_dict(include_self_link=False)
    Collection.from_dict(d)


def test_temporal_extent_init_typing() -> None:
    # This test exists purely to test the typing of the intervals argument to
    # TemporalExtent
    start_datetime = str_to_datetime("2022-01-01T00:00:00Z")
    end_datetime = str_to_datetime("2022-01-31T23:59:59Z")

    _ = TemporalExtent([[start_datetime, end_datetime]])


@pytest.mark.block_network()
def test_temporal_extent_allows_single_interval() -> None:
    start_datetime = str_to_datetime("2022-01-01T00:00:00Z")
    end_datetime = str_to_datetime("2022-01-31T23:59:59Z")

    interval = [start_datetime, end_datetime]
    temporal_extent = TemporalExtent(intervals=interval)

    assert temporal_extent.intervals == [interval]


@pytest.mark.block_network()
def test_temporal_extent_allows_single_interval_open_start() -> None:
    end_datetime = str_to_datetime("2022-01-31T23:59:59Z")

    interval = [None, end_datetime]
    temporal_extent = TemporalExtent(intervals=interval)

    assert temporal_extent.intervals == [interval]


@pytest.mark.block_network()
def test_temporal_extent_non_list_intervals_fails() -> None:
    with pytest.raises(TypeError):
        # Pass in non-list intervals
        _ = TemporalExtent(intervals=1)  # type: ignore


@pytest.mark.block_network()
def test_spatial_allows_single_bbox() -> None:
    temporal_extent = TemporalExtent(intervals=[[TEST_DATETIME, None]])

    # Pass in a single BBOX
    spatial_extent = SpatialExtent(bboxes=ARBITRARY_BBOX)

    collection_extent = Extent(spatial=spatial_extent, temporal=temporal_extent)

    collection = Collection(
        id="test", description="test desc", extent=collection_extent
    )

    # HREF required by validation
    collection.set_self_href("https://example.com/collection.json")

    collection.validate()


@pytest.mark.block_network()
def test_spatial_extent_non_list_bboxes_fails() -> None:
    with pytest.raises(TypeError):
        # Pass in non-list bboxes
        _ = SpatialExtent(bboxes=1)  # type: ignore


def test_extent_from_items() -> None:
    item1 = Item(
        id="test-item-1",
        geometry=ARBITRARY_GEOM,
        bbox=[-10, -20, 0, -10],
        datetime=datetime(2000, 2, 1, 12, 0, 0, 0, tzinfo=tz.UTC),
        properties={},
    )

    item2 = Item(
        id="test-item-2",
        geometry=ARBITRARY_GEOM,
        bbox=[0, -9, 10, 1],
        datetime=None,
        properties={
            "start_datetime": datetime_to_str(
                datetime(2000, 1, 1, 12, 0, 0, 0, tzinfo=tz.UTC)
            ),
            "end_datetime": datetime_to_str(
                datetime(2000, 7, 1, 12, 0, 0, 0, tzinfo=tz.UTC)
            ),
        },
    )

    item3 = Item(
        id="test-item-2",
        geometry=ARBITRARY_GEOM,
        bbox=[-5, -20, 5, 0],
        datetime=None,
        properties={
            "start_datetime": datetime_to_str(
                datetime(2000, 12, 1, 12, 0, 0, 0, tzinfo=tz.UTC)
            ),
            "end_datetime": datetime_to_str(
                datetime(2001, 1, 1, 12, 0, 0, 0, tzinfo=tz.UTC),
            ),
        },
    )

    extent = Extent.from_items([item1, item2, item3])
    assert len(extent.spatial.bboxes) == 1
    assert extent.spatial.bboxes[0] == [-10, -20, 10, 1]
    assert len(extent.temporal.intervals) == 1

    interval = extent.temporal.intervals[0]
    assert interval[0] == datetime(2000, 1, 1, 12, 0, 0, 0, tzinfo=tz.UTC)
    assert interval[1] == datetime(2001, 1, 1, 12, 0, 0, 0, tzinfo=tz.UTC)


def test_extent_to_from_dict() -> None:
    spatial_dict = {
        "bbox": [
            [
                172.91173669923782,
                1.3438851951615003,
                172.95469614953714,
                1.3690476620161975,
            ]
        ],
        "extension:field": "spatial value",
    }
    temporal_dict = {
        "interval": [["2020-12-11T22:38:32.125000Z", "2020-12-14T18:02:31.437000Z"]],
        "extension:field": "temporal value",
    }
    extent_dict = {
        "spatial": spatial_dict,
        "temporal": temporal_dict,
        "extension:field": "extent value",
    }
    expected_extent_extra_fields = {
        "extension:field": extent_dict["extension:field"],
    }
    expected_spatial_extra_fields = {
        "extension:field": spatial_dict["extension:field"],
    }
    expected_temporal_extra_fields = {
        "extension:field": temporal_dict["extension:field"],
    }

    extent = Extent.from_dict(extent_dict)

    assert expected_extent_extra_fields == extent.extra_fields
    assert expected_spatial_extra_fields == extent.spatial.extra_fields
    assert expected_temporal_extra_fields == extent.temporal.extra_fields

    assert extent_dict == extent.to_dict()


class TestCollectionSubClass:
    """This tests cases related to creating classes inheriting from pystac.Catalog to
    ensure that inheritance, class methods, etc. function as expected."""

    MULTI_EXTENT = TestCases.get_path("data-files/collections/multi-extent.json")

    class BasicCustomCollection(pystac.Collection):
        def get_items(self) -> Iterator[Item]:  # type: ignore
            # This get_items does not have the `recursive` kwarg. This mimics
            # the current state of pystac-client and is intended to test
            # backwards compatibility of inherited classes
            return super().get_items()

    def test_from_dict_returns_subclass(self) -> None:
        stac_io = pystac.StacIO.default()
        collection_dict = stac_io.read_json(self.MULTI_EXTENT)
        custom_collection = self.BasicCustomCollection.from_dict(collection_dict)

        assert isinstance(custom_collection, self.BasicCustomCollection)

    def test_from_file_returns_subclass(self) -> None:
        custom_collection = self.BasicCustomCollection.from_file(self.MULTI_EXTENT)

        assert isinstance(custom_collection, self.BasicCustomCollection)

    def test_clone(self) -> None:
        custom_collection = self.BasicCustomCollection.from_file(self.MULTI_EXTENT)
        cloned_collection = custom_collection.clone()

        assert isinstance(cloned_collection, self.BasicCustomCollection)

    def test_collection_get_item_works(self) -> None:
        path = TestCases.get_path(
            "data-files/catalogs/test-case-1/country-1/area-1-1/collection.json"
        )
        custom_collection = self.BasicCustomCollection.from_file(path)
        collection = custom_collection.clone()
        with pytest.warns(DeprecationWarning):
            collection.get_item("area-1-1-imagery")


def test_collection_get_item_raises_type_error() -> None:
    class BasicCustomCollection(pystac.Collection):
        def get_items(  # type: ignore
            self, *, recursive: bool = False
        ) -> Iterator[Item]:
            # This get_items does not allow ids as args.
            return super().get_items(recursive=recursive)

    path = TestCases.get_path(
        "data-files/catalogs/test-case-1/country-1/area-1-1/collection.json"
    )
    custom_collection = BasicCustomCollection.from_file(path)
    collection = custom_collection.clone()
    with pytest.raises(TypeError, match="takes 1 positional argument"):
        collection.get_item("area-1-1-imagery")


def test_custom_collection_from_dict(collection: Collection) -> None:
    # https://github.com/stac-utils/pystac/issues/862
    class CustomCollection(Collection):
        @classmethod
        def from_dict(
            cls,
            d: dict[str, Any],
            href: str | None = None,
            root: Catalog | None = None,
            migrate: bool = True,
            preserve_dict: bool = True,
        ) -> CustomCollection:
            return super().from_dict(d)

    _ = CustomCollection.from_dict(collection.to_dict())


@pytest.mark.parametrize("add_canonical", (True, False))
def test_remove_hierarchical_links(
    test_case_1_catalog: Catalog, add_canonical: bool
) -> None:
    collection = list(test_case_1_catalog.get_all_collections())[0]
    collection.remove_hierarchical_links(add_canonical=add_canonical)
    for link in collection.links:
        assert not link.is_hierarchical()
    assert bool(collection.get_single_link("canonical")) == add_canonical


@pytest.mark.parametrize("child", ["country-1", "country-2"])
def test_get_child_checks_links_where_hrefs_contains_id_first(
    test_case_1_catalog: Catalog, child: str
) -> None:
    cat1 = test_case_1_catalog
    country = cat1.get_child(child)
    assert country is not None
    child_links = [link for link in cat1.links if link.rel == pystac.RelType.CHILD]
    for link in child_links:
        if country.id not in link.href:
            assert not link.is_resolved()


def test_get_child_sort_links_by_id_is_configurable(
    test_case_1_catalog: Catalog,
) -> None:
    cat1 = test_case_1_catalog
    country = cat1.get_child("country-2", sort_links_by_id=False)
    assert country is not None
    child_links = [link for link in cat1.links if link.rel == pystac.RelType.CHILD]
    for link in child_links:
        assert link.is_resolved()


def test_get_item_returns_none_if_not_found(
    test_case_8_collection: Collection,
) -> None:
    col8 = test_case_8_collection
    item = col8.get_item("notarealitem")
    assert item is None


def test_get_item_is_not_recursive_by_default(
    test_case_8_collection: Collection,
) -> None:
    col8 = test_case_8_collection
    item = col8.get_item("20170831_162740_ssc1d1")
    assert item is None

    item = col8.get_item("20170831_162740_ssc1d1", recursive=True)
    assert item is not None


def test_delete_asset(tmp_asset: Asset, collection: Collection) -> None:
    asset = tmp_asset
    href = asset.get_absolute_href()
    item = asset.owner
    name = "foo"

    assert href is not None
    assert item is not None

    collection.add_asset(name, asset)

    # steal the href from the owner and use it as the collection's
    collection.set_self_href(item.get_self_href())

    collection.delete_asset(name)
    assert name not in collection.assets
    assert not os.path.exists(href)


def test_delete_asset_relative_no_self_link_fails(
    tmp_asset: Asset, collection: Collection
) -> None:
    asset = tmp_asset
    href = asset.get_absolute_href()
    name = "foo"

    assert href is not None

    collection.add_asset(name, asset)

    with pytest.raises(ValueError, match="Cannot delete file") as e:
        collection.delete_asset(name)

    assert asset.href in str(e.value)
    assert name in collection.assets
    assert os.path.exists(href)


def test_permissive_temporal_extent_deserialization(collection: Collection) -> None:
    # https://github.com/stac-utils/pystac/issues/1221
    collection_dict = collection.to_dict(include_self_link=False, transform_hrefs=False)
    collection_dict["extent"]["temporal"]["interval"] = collection_dict["extent"][
        "temporal"
    ]["interval"][0]
    with pytest.warns(UserWarning):
        Collection.from_dict(collection_dict)


@pytest.mark.parametrize("fixture_name", ("sample_item_collection", "sample_items"))
def test_from_items(fixture_name: str, request: pytest.FixtureRequest) -> None:
    items = request.getfixturevalue(fixture_name)
    collection = Collection.from_items(items)

    for item in items:
        assert collection.id == item.collection_id
        assert collection.extent.spatial.bboxes[0][0] <= item.bbox[0]
        assert collection.extent.spatial.bboxes[0][1] <= item.bbox[1]
        assert collection.extent.spatial.bboxes[0][2] >= item.bbox[2]
        assert collection.extent.spatial.bboxes[0][3] >= item.bbox[3]

        start = collection.extent.temporal.intervals[0][0]
        end = collection.extent.temporal.intervals[0][1]
        assert start and start <= str_to_datetime(item.properties["start_datetime"])
        assert end and end >= str_to_datetime(item.properties["end_datetime"])

    if isinstance(items, ItemCollection):
        expected = {(link["rel"], link["href"]) for link in items.extra_fields["links"]}
        actual = {(link.rel, link.href) for link in collection.links}
        assert expected.issubset(actual)


def test_from_items_pulls_from_properties() -> None:
    item1 = Item(
        id="test-item-1",
        geometry=ARBITRARY_GEOM,
        bbox=[-10, -20, 0, -10],
        datetime=datetime(2000, 2, 1, 12, 0, 0, 0, tzinfo=tz.UTC),
        collection="test-collection-1",
        properties={"title": "Test Item", "description": "Extra words describing"},
    )
    collection = Collection.from_items([item1])
    assert collection.id == item1.collection_id
    assert collection.title == item1.properties["title"]
    assert collection.description == item1.properties["description"]


def test_from_items_without_collection_id() -> None:
    item1 = Item(
        id="test-item-1",
        geometry=ARBITRARY_GEOM,
        bbox=[-10, -20, 0, -10],
        datetime=datetime(2000, 2, 1, 12, 0, 0, 0, tzinfo=tz.UTC),
        properties={},
    )
    with pytest.raises(ValueError, match="Collection id must be defined."):
        Collection.from_items([item1])

    collection = Collection.from_items([item1], id="test-collection")
    assert collection.id == "test-collection"


def test_from_items_with_collection_ids() -> None:
    item1 = Item(
        id="test-item-1",
        geometry=ARBITRARY_GEOM,
        bbox=[-10, -20, 0, -10],
        datetime=datetime(2000, 2, 1, 12, 0, 0, 0, tzinfo=tz.UTC),
        collection="test-collection-1",
        properties={},
    )
    item2 = Item(
        id="test-item-2",
        geometry=ARBITRARY_GEOM,
        bbox=[-15, -20, 0, -10],
        datetime=datetime(2000, 2, 1, 13, 0, 0, 0, tzinfo=tz.UTC),
        collection="test-collection-2",
        properties={},
    )

    with pytest.raises(ValueError, match="Collection id must be defined."):
        Collection.from_items([item1, item2])

    collection = Collection.from_items([item1, item2], id="test-collection")
    assert collection.id == "test-collection"


def test_from_items_with_different_values() -> None:
    item1 = Item(
        id="test-item-1",
        geometry=ARBITRARY_GEOM,
        bbox=[-10, -20, 0, -10],
        datetime=datetime(2000, 2, 1, 12, 0, 0, 0, tzinfo=tz.UTC),
        properties={"title": "Test Item 1"},
    )
    item2 = Item(
        id="test-item-2",
        geometry=ARBITRARY_GEOM,
        bbox=[-15, -20, 0, -10],
        datetime=datetime(2000, 2, 1, 13, 0, 0, 0, tzinfo=tz.UTC),
        properties={"title": "Test Item 2"},
    )

    collection = Collection.from_items([item1, item2], id="test_collection")
    assert collection.title is None


def test_from_items_with_providers(sample_item_collection: ItemCollection) -> None:
    sample_item_collection.extra_fields["providers"] = [{"name": "pystac"}]

    collection = Collection.from_items(sample_item_collection)
    assert collection.providers and len(collection.providers) == 1

    provider = collection.providers[0]
    assert provider and provider.name == "pystac"


def test_from_dict_null_extent(collection: Collection) -> None:
    # https://github.com/stac-utils/pystac/issues/1558
    # https://github.com/EOPF-Sample-Service/eopf-stac/issues/18
    d = collection.to_dict()
    d["extent"] = None
    with pytest.warns(UserWarning):
        c = Collection.from_dict(d)

    assert c.extent.spatial.to_dict()["bbox"] == [[-90, -180, 90, 180]]
    assert c.extent.temporal.to_dict()["interval"] == [[None, None]]


def test_from_dict_missing_extent(collection: Collection) -> None:
    d = collection.to_dict()
    del d["extent"]
    with pytest.warns(UserWarning):
        c = Collection.from_dict(d)

    assert c.extent.spatial.to_dict()["bbox"] == [[-90, -180, 90, 180]]
    assert c.extent.temporal.to_dict()["interval"] == [[None, None]]



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Iterable
from enum import Enum
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Protocol,
    TypeVar,
)

import pystac
from pystac.utils import get_required

if TYPE_CHECKING:
    from pystac.collection import Collection
    from pystac.item import Item


def __getattr__(name: str) -> Any:
    if name == "FIELDS_JSON_URL":
        import warnings

        warnings.warn(
            "FIELDS_JSON_URL is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return (
            "https://cdn.jsdelivr.net/npm/@radiantearth/"
            "stac-fields/fields-normalized.json"
        )
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class _Comparable_x(Protocol):
    """Protocol for annotating comparable types.

    For matching __lt__ that takes an 'x' parameter
    (e.g. float)
    """

    @abstractmethod
    def __lt__(self: T, x: T) -> bool:
        return NotImplemented


class _Comparable_other(Protocol):
    """Protocol for annotating comparable types.

    For matching __lt___ that takes an 'other' parameter
    (e.g. datetime)
    """

    @abstractmethod
    def __lt__(self: T, other: T) -> bool:
        return NotImplemented


T = TypeVar("T", bound=_Comparable_x | _Comparable_other)


class RangeSummary(Generic[T]):
    minimum: T
    maximum: T

    def __init__(self, minimum: T, maximum: T):
        self.minimum = minimum
        self.maximum = maximum

    def to_dict(self) -> dict[str, Any]:
        return {"minimum": self.minimum, "maximum": self.maximum}

    def update_with_value(self, v: T) -> None:
        self.minimum = min(self.minimum, v)
        self.maximum = max(self.maximum, v)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> RangeSummary[T]:
        minimum: T = get_required(d.get("minimum"), "RangeSummary", "minimum")
        maximum: T = get_required(d.get("maximum"), "RangeSummary", "maximum")
        return cls(minimum=minimum, maximum=maximum)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, RangeSummary):
            return NotImplemented

        return self.to_dict() == o.to_dict()

    def __repr__(self) -> str:
        return self.to_dict().__repr__()


@lru_cache(maxsize=None)
def _get_fields_json(url: str | None) -> dict[str, Any]:
    if url is None:
        import importlib.resources
        import json

        # Every time pystac is released this file gets pulled from
        # https://cdn.jsdelivr.net/npm/@radiantearth/stac-fields/fields-normalized.json
        jsonfields: dict[str, Any] = json.loads(
            importlib.resources.files("pystac.static")
            .joinpath("fields-normalized.json")
            .read_text()
        )
        return jsonfields
    return pystac.StacIO.default().read_json(url)


class SummaryStrategy(Enum):
    ARRAY = "v"
    RANGE = "r"
    SCHEMA = "s"
    DONT_SUMMARIZE = False
    DEFAULT = True


class Summarizer:
    """The Summarizer computes summaries from values, following the definition of fields
    to summarize.

    The fields to summarize can be provided as a JSON file or as a dictionary of
    field names and SummaryStrategys. If nothing is provided, a default JSON file
    will be used.

    Only fields that are in the Item `properties` can be summarized.
    Thus, it is not possible to summarize the top-level fields such as `id` or `assets`.

    For more information about the structure of the fields JSON file, see:
    https://github.com/stac-utils/stac-fields

    The default JSON file used is a snapshot of the following file at the time of
    the pystac release:
    https://cdn.jsdelivr.net/npm/@radiantearth/stac-fields/fields-normalized.json

    Args:
        fields: A string containing the path to the JSON file with field descriptions.
            Alternatively, a dict with the field names as keys and SummaryStrategys
            as values.
            If nothing is passed, a default file with field descriptions will be used.
    """

    summaryfields: dict[str, SummaryStrategy]

    def __init__(self, fields: str | dict[str, SummaryStrategy] | None = None):
        if isinstance(fields, dict):
            self._set_field_definitions(fields)
        else:
            jsonfields = _get_fields_json(fields)
            self._set_field_definitions(jsonfields["metadata"])

    def _set_field_definitions(self, fields: dict[str, Any]) -> None:
        self.summaryfields = {}
        for name, desc in fields.items():
            strategy: SummaryStrategy = SummaryStrategy.DEFAULT
            if isinstance(desc, SummaryStrategy):
                strategy = desc
            elif isinstance(desc, dict):
                strategy_value = desc.get("summary", True)
                try:
                    strategy = SummaryStrategy(strategy_value)
                except ValueError:
                    pass

            if strategy != SummaryStrategy.DONT_SUMMARIZE:
                self.summaryfields[name] = strategy

    def _update_with_item(self, summaries: Summaries, item: Item) -> None:
        import numbers

        for k, v in item.properties.items():
            if k in self.summaryfields:
                strategy = self.summaryfields[k]
                if strategy == SummaryStrategy.RANGE or (
                    strategy == SummaryStrategy.DEFAULT
                    and isinstance(v, numbers.Number)
                    and not isinstance(v, bool)
                ):
                    rangesummary: RangeSummary[Any] | None = summaries.get_range(k)
                    if rangesummary is None:
                        summaries.add(k, RangeSummary(v, v))
                    else:
                        rangesummary.update_with_value(v)
                elif strategy == SummaryStrategy.ARRAY or (
                    strategy == SummaryStrategy.DEFAULT and isinstance(v, list)
                ):
                    listsummary: list[Any] = summaries.get_list(k) or []
                    if not isinstance(v, list):
                        v = [v]
                    for element in v:
                        if element not in listsummary:
                            listsummary.append(element)
                    summaries.add(k, listsummary)
                else:
                    summary: list[Any] = summaries.get_list(k) or []
                    if v not in summary:
                        summary.append(v)
                    summaries.add(k, summary)

    def summarize(self, source: Collection | Iterable[Item]) -> Summaries:
        """Creates summaries from items"""
        summaries = Summaries.empty()
        if isinstance(source, pystac.Collection):
            for item in source.get_items(recursive=True):
                self._update_with_item(summaries, item)
        else:
            for item in source:
                self._update_with_item(summaries, item)

        return summaries


DEFAULT_MAXCOUNT = 25


class Summaries:
    _summaries: dict[str, Any]

    lists: dict[str, list[Any]]
    other: dict[str, Any]
    ranges: dict[str, RangeSummary[Any]]
    schemas: dict[str, dict[str, Any]]
    maxcount: int

    def __init__(
        self, summaries: dict[str, Any], maxcount: int = DEFAULT_MAXCOUNT
    ) -> None:
        self._summaries = summaries
        self.maxcount = maxcount

        self.lists = {}
        self.ranges = {}
        self.schemas = {}
        self.other = {}

        for prop_key, summary in summaries.items():
            self.add(prop_key, summary)

    def get_list(self, prop: str) -> list[Any] | None:
        return self.lists.get(prop)

    def get_range(self, prop: str) -> RangeSummary[Any] | None:
        return self.ranges.get(prop)

    def get_schema(self, prop: str) -> dict[str, Any] | None:
        return self.schemas.get(prop)

    def add(
        self,
        prop_key: str,
        summary: list[Any] | RangeSummary[Any] | dict[str, Any],
    ) -> None:
        if isinstance(summary, list):
            self.lists[prop_key] = summary
        elif isinstance(summary, dict):
            if "minimum" in summary:
                self.ranges[prop_key] = RangeSummary[Any].from_dict(summary)
            else:
                self.schemas[prop_key] = summary
        elif isinstance(summary, RangeSummary):
            self.ranges[prop_key] = summary
        else:
            self.other[prop_key] = summary

    def remove(self, prop_key: str) -> None:
        self.lists.pop(prop_key, None)
        self.ranges.pop(prop_key, None)
        self.schemas.pop(prop_key, None)
        self.other.pop(prop_key, None)

    def update(self, summaries: Summaries) -> None:
        self.lists.update(summaries.lists)
        self.ranges.update(summaries.ranges)
        self.schemas.update(summaries.schemas)
        self.other.update(summaries.other)

    def combine(self, summaries: Summaries) -> None:
        for listname, listvalue in summaries.lists.items():
            if listname in self.lists:
                self.lists[listname].extend(listvalue)
            else:
                self.lists[listname] = listvalue
        for rangename, rang in summaries.ranges.items():
            if rangename in self.ranges:
                self.ranges[rangename].update_with_value(rang.minimum)
                self.ranges[rangename].update_with_value(rang.maximum)
            else:
                self.ranges[rangename] = rang
        for schemaname, schema in summaries.schemas.items():
            if schemaname in self.schemas:
                self.schemas[schemaname].update(schema)
            else:
                self.schemas[schemaname] = schema
        for k, v in summaries.other.items():
            if k in self.other:
                self.other[k].update(v)
            else:
                self.other[k] = v

    def is_empty(self) -> bool:
        return not (
            any(self.lists) or any(self.ranges) or any(self.schemas) or any(self.other)
        )

    def clone(self) -> Summaries:
        """Clones this object.

        Returns:
            Summaries: The clone of this object
        """
        from copy import deepcopy

        cls = self.__class__
        summaries = cls(summaries=deepcopy(self._summaries), maxcount=self.maxcount)
        summaries.lists = deepcopy(self.lists)
        summaries.other = deepcopy(self.other)
        summaries.ranges = deepcopy(self.ranges)
        summaries.schemas = deepcopy(self.schemas)
        return summaries

    def to_dict(self) -> dict[str, Any]:
        return {
            **{k: v for k, v in self.lists.items() if len(v) < self.maxcount},
            **{k: v.to_dict() for k, v in self.ranges.items()},
            **self.schemas,
            **self.other,
        }

    @classmethod
    def empty(cls) -> Summaries:
        return Summaries({})



from __future__ import annotations

from collections.abc import Collection, Iterable, Iterator
from typing import (
    Any,
    TypeAlias,
    TypeVar,
)

import pystac
from pystac.errors import STACTypeError
from pystac.serialization.identify import identify_stac_object_type
from pystac.utils import HREF, is_absolute_href, make_absolute_href, make_posix_style

ItemLike: TypeAlias = pystac.Item | dict[str, Any]

#: Generalized version of :class:`ItemCollection`
C = TypeVar("C", bound="ItemCollection")


class ItemCollection(Collection[pystac.Item]):
    """Implementation of a GeoJSON FeatureCollection whose features are all STAC
    Items,
    as defined by `STAC API - ItemCollection Fragment <https://github.com/radiantearth/stac-api-spec/blob/release/v1.0.0/fragments/itemcollection/README.md>`_.

    All :class:`~pystac.Item` instances passed to the :class:`ItemCollection` instance
    during instantiation are cloned and have their ``"root"`` URL cleared. Instances of
    this class implement the abstract methods of :class:`typing.Collection` and can also
    be added together (see below for examples using these methods).

    Any additional top-level fields in the FeatureCollection are retained in
    :attr:`~pystac.ItemCollection.extra_fields` by the
    :meth:`~pystac.ItemCollection.from_dict` and
    :meth:`~pystac.ItemCollection.from_file`
    methods and will be present in the serialized file
    from :meth:`~pystac.ItemCollection.save_object`.

    Arguments:

        items : List of :class:`~pystac.Item` instances to include in the
            :class:`ItemCollection`.
        extra_fields : Dictionary of additional top-level fields included in the
            :class:`ItemCollection`.
        clone_items : Optional flag indicating whether :class:`~pystac.Item` instances
            should be cloned before storing in the :class:`ItemCollection`. Setting to
            ``False`` will result in faster instantiation, but changes made to
            :class:`~pystac.Item` instances in the :class:`ItemCollection` will mutate
            the original ``Item``. Defaults to ``True``.

    Examples:

        Loop over all items in the :class`ItemCollection`

        >>> item_collection: ItemCollection = ...
        >>> for item in item_collection:
        ...     ...

        Get the number of :class:`~pystac.Item` instances in the
        :class:`ItemCollection`

        >>> length: int = len(item_collection)

        Check if an :class:`~pystac.Item` is in the :class:`ItemCollection`. Note
        that the ``clone_items`` argument must be ``False`` for this to return
        ``True``, since equality of PySTAC objects is currently evaluated using default
        object equality (i.e. ``item_1 is item_2``).

        >>> item: Item = ...
        >>> item_collection = ItemCollection(items=[item], clone_items=False)
        >>> assert item in item_collection

        Combine :class:`ItemCollection` instances

        >>> item_1: Item = ...
        >>> item_2: Item = ...
        >>> item_3: Item = ...
        >>> item_collection_1 = ItemCollection(items=[item_1, item_2])
        >>> item_collection_2 = ItemCollection(items=[item_2, item_3])
        >>> combined = item_collection_1 + item_collection_2
        >>> assert len(combined) == 4
        # If an item is present in both ItemCollections it will occur twice
    """

    items: list[pystac.Item]
    """List of :class:`pystac.Item` instances contained in this ``ItemCollection``."""

    extra_fields: dict[str, Any]
    """Dictionary of additional top-level fields for the GeoJSON
    FeatureCollection."""

    def __init__(
        self,
        items: Iterable[ItemLike],
        extra_fields: dict[str, Any] | None = None,
        clone_items: bool = True,
    ):
        def map_item(item_or_dict: ItemLike) -> pystac.Item:
            # Converts dicts to pystac.Items and clones if necessary
            if isinstance(item_or_dict, pystac.Item):
                return item_or_dict.clone() if clone_items else item_or_dict
            else:
                return pystac.Item.from_dict(item_or_dict, preserve_dict=clone_items)

        self.items = list(map(map_item, items))
        self.extra_fields = extra_fields or {}

    def __getitem__(self, idx: int) -> pystac.Item:
        return self.items[idx]

    def __iter__(self) -> Iterator[pystac.Item]:
        return iter(self.items)

    def __len__(self) -> int:
        return len(self.items)

    def __contains__(self, __x: object) -> bool:
        return __x in self.items

    def __add__(self, other: object) -> ItemCollection:
        if not isinstance(other, ItemCollection):
            return NotImplemented

        combined = [*self.items, *other.items]
        return ItemCollection(items=combined)

    def to_dict(self, transform_hrefs: bool = False) -> dict[str, Any]:
        """Serializes an :class:`ItemCollection` instance to a dictionary.

        Args:
            transform_hrefs: If True, transform the HREF of hierarchical links
                of Items based on the type of catalog the Item belongs to (if any).
                I.e. if the item belongs to a root catalog that is
                RELATIVE_PUBLISHED or SELF_CONTAINED,
                hierarchical link HREFs will be transformed to be relative to the
                catalog root. This can be slow if the Items have root links that
                have not yet been resolved. Defaults to False.
        """
        return {
            "type": "FeatureCollection",
            "features": [
                item.to_dict(transform_hrefs=transform_hrefs) for item in self.items
            ],
            **self.extra_fields,
        }

    def _repr_html_(self) -> str:
        from html import escape

        from pystac.html.jinja_env import get_jinja_env

        jinja_env = get_jinja_env()
        if jinja_env:
            template = jinja_env.get_template("JSON.jinja2")
            return str(template.render(dict=self.to_dict(), plain=escape(repr(self))))
        else:
            return escape(repr(self))

    def clone(self) -> ItemCollection:
        """Creates a clone of this instance. This clone is a deep copy; all
        :class:`~pystac.Item` instances are cloned and all additional top-level fields
        are deep copied."""
        from copy import deepcopy

        return self.__class__(
            items=[item.clone() for item in self.items],
            extra_fields=deepcopy(self.extra_fields),
        )

    @classmethod
    def from_dict(
        cls: type[C],
        d: dict[str, Any],
        preserve_dict: bool = True,
        root: pystac.Catalog | None = None,
    ) -> C:
        """Creates a :class:`ItemCollection` instance from a dictionary.

        Arguments:
            d : The dictionary from which the :class:`ItemCollection` will be created
            preserve_dict: If False, the dict parameter ``d`` may be modified
                during this method call. Otherwise, the dict is not mutated.
                Defaults to True, which results in a deepcopy of the
                parameter. Set to False when possible to avoid the performance
                hit of a deepcopy.
        """
        if not cls.is_item_collection(d):
            raise STACTypeError(d, cls)

        items = [
            pystac.Item.from_dict(item, preserve_dict=preserve_dict, root=root)
            for item in d.get("features", [])
        ]
        extra_fields = {k: v for k, v in d.items() if k not in ("features", "type")}

        return cls(items=items, extra_fields=extra_fields)

    @classmethod
    def from_file(cls: type[C], href: HREF, stac_io: pystac.StacIO | None = None) -> C:
        """Reads a :class:`ItemCollection` from a JSON file.

        Arguments:
            href : Path to the file.
            stac_io : A :class:`~pystac.StacIO` instance to use for file I/O
        """
        if stac_io is None:
            stac_io = pystac.StacIO.default()

        href = make_posix_style(href)
        if not is_absolute_href(href):
            href = make_absolute_href(href)

        d = stac_io.read_json(href)

        return cls.from_dict(d, preserve_dict=False)

    def save_object(
        self,
        dest_href: str,
        stac_io: pystac.StacIO | None = None,
    ) -> None:
        """Saves this instance to the ``dest_href`` location.

        Args:
            dest_href : Location to which the file will be saved.
            stac_io: Optional :class:`~pystac.StacIO` instance to use. If not provided,
                will use the default instance.
        """
        if stac_io is None:
            stac_io = pystac.StacIO.default()

        stac_io.save_json(dest_href, self.to_dict())

    @staticmethod
    def is_item_collection(d: dict[str, Any]) -> bool:
        """Checks if the given dictionary represents a valid :class:`ItemCollection`.

        Args:
            d : Dictionary to check
        """
        typ = d.get("type")

        # All ItemCollections are GeoJSON FeatureCollections
        if typ != "FeatureCollection":
            return False

        # If it is a FeatureCollection and has a "stac_version" field, then it is an
        #  ItemCollection. This will cover ItemCollections from STAC 0.9 to
        #  <1.0.0-beta.1, when ItemCollections were removed from the core STAC Spec
        if "stac_version" in d:
            return True

        # Prior to STAC 0.9 ItemCollections did not have a stac_version field and could
        #  only be identified by the fact that all of their 'features' are STAC Items.
        return all(
            identify_stac_object_type(feature) == pystac.STACObjectType.ITEM
            for feature in d.get("features", [])
        )



"""Implements the :stac-ext:`Label Extension <label>`."""

from __future__ import annotations

import warnings
from collections.abc import Iterable, Sequence
from typing import Any, Literal, cast

import pystac
from pystac.extensions.base import ExtensionManagementMixin, SummariesExtension
from pystac.extensions.hooks import ExtensionHooks
from pystac.serialization.identify import STACJSONDescription, STACVersionID
from pystac.utils import StringEnum, get_required, map_opt

warnings.warn(
    "The PySTAC Label Extension is deprecated. The extension itself "
    '(https://github.com/stac-extensions/label) is currently unmaintained, in "pilot" '
    "maturity, and has significant issues.",
    DeprecationWarning,
)

SCHEMA_URI = "https://stac-extensions.github.io/label/v1.0.1/schema.json"
SCHEMA_URIS = [
    "https://stac-extensions.github.io/label/v1.0.0/schema.json",
    SCHEMA_URI,
]
PREFIX = "label:"

PROPERTIES_PROP = PREFIX + "properties"
CLASSES_PROP = PREFIX + "classes"
DESCRIPTION_PROP = PREFIX + "description"
TYPE_PROP = PREFIX + "type"
TASKS_PROP = PREFIX + "tasks"
METHODS_PROP = PREFIX + "methods"
OVERVIEWS_PROP = PREFIX + "overviews"


class LabelRelType(StringEnum):
    """A list of rel types defined in the Label Extension.

    See the :stac-ext:`Label Extension Links <label#links-source-imagery>`
    documentation for details.
    """

    SOURCE = "source"
    """Used to indicate a link to the source item to which a label item applies."""


class LabelType(StringEnum):
    """Enumerates valid label types ("raster" or "vector")."""

    VECTOR = "vector"
    RASTER = "raster"

    ALL = [VECTOR, RASTER]
    """Convenience attribute for checking if values are valid label types"""


class LabelTask(StringEnum):
    """Enumerates recommended values for "label:tasks" field."""

    REGRESSION = "regression"
    CLASSIFICATION = "classification"
    DETECTION = "detection"
    SEGMENTATION = "segmentation"


class LabelMethod(StringEnum):
    """Enumerates recommended values for "label:methods" field."""

    AUTOMATED = "automated"
    MANUAL = "manual"


class LabelClasses:
    """Defines the list of possible class names (e.g., tree, building, car, hippo).

    Use :meth:`LabelClasses.create` to create a new instance from property values.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def apply(
        self,
        classes: Sequence[str | int | float],
        name: str | None = None,
    ) -> None:
        """Sets the properties for this instance.

        Args:
            classes : The different possible class values.
            name : The property key within the asset's each Feature corresponding
                to class labels. If labels are raster-formatted, do not supply;
                required otherwise.
        """
        self.classes = classes
        self.name = name

    @classmethod
    def create(
        cls,
        classes: Sequence[str | int | float],
        name: str | None = None,
    ) -> LabelClasses:
        """Creates a new :class:`~LabelClasses` instance.

        Args:
            classes : The different possible class values.
            name : The property key within the asset's each Feature corresponding
                to class labels. If labels are raster-formatted, do not supply;
                required otherwise.
        """
        c = cls({})
        c.apply(classes, name)
        return c

    @property
    def classes(self) -> Sequence[str | int | float]:
        """Gets or sets the class values."""
        return get_required(self.properties.get("classes"), self, "classes")

    @classes.setter
    def classes(self, v: Sequence[str | int | float]) -> None:
        self.properties["classes"] = v

    @property
    def name(self) -> str | None:
        """Gets or sets the property key within each Feature in the asset corresponding
        to class labels. If labels are raster-formatted, use ``None``.
        """
        return self.properties.get("name")

    @name.setter
    def name(self, v: str | None) -> None:
        # The "name" property is required but may be null
        self.properties["name"] = v

    def __repr__(self) -> str:
        return "<ClassObject classes={}>".format(
            ",".join([str(x) for x in self.classes])
        )

    def __eq__(self, o: object) -> bool:
        if isinstance(o, LabelClasses):
            o = o.to_dict()

        if not isinstance(o, dict):
            return NotImplemented

        return self.to_dict() == o

    def to_dict(self) -> dict[str, Any]:
        """Returns this label classes object as a dictionary."""
        return self.properties


class LabelCount:
    """Contains counts for categorical data.

    Use :meth:`LabelCount.create` to create a new instance.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def apply(self, name: str, count: int) -> None:
        """Sets the properties for this instance.

        Args:
            name : One of the different possible classes within the property.
            count : The number of occurrences of the class.
        """
        self.name = name
        self.count = count

    @classmethod
    def create(cls, name: str, count: int) -> LabelCount:
        """Creates a :class:`LabelCount` instance.

        Args:
            name : One of the different possible classes within the property.
            count : The number of occurrences of the class.
        """
        x = cls({})
        x.apply(name, count)
        return x

    @property
    def name(self) -> str:
        """Gets or sets the class that this count represents."""
        return cast(str, get_required(self.properties.get("name"), self, "name"))

    @name.setter
    def name(self, v: str) -> None:
        self.properties["name"] = v

    @property
    def count(self) -> int:
        """Get or sets the number of occurrences of the class."""
        return cast(int, get_required(self.properties.get("count"), self, "count"))

    @count.setter
    def count(self, v: int) -> None:
        self.properties["count"] = v

    def to_dict(self) -> dict[str, Any]:
        """Returns this label count object as a dictionary."""
        return self.properties

    def __eq__(self, o: object) -> bool:
        if isinstance(o, LabelCount):
            o = o.to_dict()

        if not isinstance(o, dict):
            return NotImplemented

        return self.to_dict() == o


class LabelStatistics:
    """Contains statistics for regression/continuous numeric value data.

    Use :meth:`LabelStatistics.create` to create a new instance.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    def apply(self, name: str, value: float) -> None:
        """Sets the property values for this instance.

        Args:
            name : The name of the statistic being reported.
            value : The value of the statistic
        """
        self.name = name
        self.value = value

    @classmethod
    def create(cls, name: str, value: float) -> LabelStatistics:
        """Creates a new :class:`LabelStatistics` instance.

        Args:
            name : The name of the statistic being reported.
            value : The value of the statistic
        """
        x = cls({})
        x.apply(name, value)
        return x

    @property
    def name(self) -> str:
        """Gets or sets the name of the statistic being reported."""
        return cast(str, get_required(self.properties.get("name"), self, "name"))

    @name.setter
    def name(self, v: str) -> None:
        self.properties["name"] = v

    @property
    def value(self) -> float:
        """Gets or sets the value of the statistic."""
        return cast(float, get_required(self.properties.get("value"), self, "value"))

    @value.setter
    def value(self, v: float) -> None:
        self.properties["value"] = v

    def to_dict(self) -> dict[str, Any]:
        """Returns this label statistics object as a dictionary."""
        return self.properties

    def __eq__(self, o: object) -> bool:
        if isinstance(o, LabelStatistics):
            o = o.to_dict()

        if not isinstance(o, dict):
            return NotImplemented

        return self.to_dict() == o


class LabelOverview:
    """Stores counts (for classification-type data) or summary statistics (for
    continuous numerical/regression data).

    Use :meth:`LabelOverview.create` to create a new instance.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def apply(
        self,
        property_key: str | None,
        counts: list[LabelCount] | None = None,
        statistics: list[LabelStatistics] | None = None,
    ) -> None:
        """Sets the properties for this instance.

        Either ``counts`` or ``statistics``, or both, can be placed in an overview;
        at least one is required.

        Args:
            property_key : The property key within the asset corresponding to
                class labels that these counts or statistics are referencing. If the
                label data is raster data, this should be None.
            counts: Optional list of :class:`LabelCounts` containing counts
                for categorical data.
            statistics: Optional list of :class:`LabelStatistics` containing statistics
                for regression/continuous numeric value data.
        """
        self.property_key = property_key
        self.counts = counts
        self.statistics = statistics

    @classmethod
    def create(
        cls,
        property_key: str | None,
        counts: list[LabelCount] | None = None,
        statistics: list[LabelStatistics] | None = None,
    ) -> LabelOverview:
        """Creates a new instance.

        Either ``counts`` or ``statistics``, or both, can be placed in an overview;
        at least one is required.

        Args:
            property_key : The property key within the asset corresponding to
                class labels.
            counts: Optional list of :class:`LabelCounts` containing counts for
                categorical data.
            statistics: Optional list of :class:`LabelStatistics` containing statistics
                for regression/continuous numeric value data.
        """
        x = LabelOverview({})
        x.apply(property_key, counts, statistics)
        return x

    @property
    def property_key(self) -> str | None:
        """Gets or sets the property key within the asset corresponding to class
        labels."""
        return self.properties.get("property_key")

    @property_key.setter
    def property_key(self, v: str | None) -> None:
        self.properties["property_key"] = v

    @property
    def counts(self) -> list[LabelCount] | None:
        """Gets or sets the list of :class:`LabelCounts` containing counts for
        categorical data."""
        counts = self.properties.get("counts")
        if counts is None:
            return None
        return [LabelCount(c) for c in counts]

    @counts.setter
    def counts(self, v: list[LabelCount] | None) -> None:
        if v is None:
            self.properties.pop("counts", None)
        else:
            if not isinstance(v, list):
                raise pystac.STACError(f"counts must be a list! Invalid input: {v}")

            self.properties["counts"] = [c.to_dict() for c in v]

    @property
    def statistics(self) -> list[LabelStatistics] | None:
        """Gets or sets the list of :class:`LabelStatistics` containing statistics for
        regression/continuous numeric value data."""
        statistics = self.properties.get("statistics")
        if statistics is None:
            return None

        return [LabelStatistics(s) for s in statistics]

    @statistics.setter
    def statistics(self, v: list[LabelStatistics] | None) -> None:
        if v is None:
            self.properties.pop("statistics", None)
        else:
            self.properties["statistics"] = [s.to_dict() for s in v]

    def merge_counts(self, other: LabelOverview) -> LabelOverview:
        """Merges the counts associated with this overview with another overview.
        Creates a new instance.

        Args:
            other : The other LabelOverview to merge.

        Returns:
            A new LabelOverview with the counts merged. This will
            drop any statistics associated with either of the LabelOverviews.
        """
        assert self.property_key == other.property_key

        if self.counts is None:
            new_counts = other.counts
        else:
            if other.counts is None:
                new_counts = self.counts
            else:
                count_by_prop: dict[str, int] = {}

                def add_counts(counts: list[LabelCount]) -> None:
                    for c in counts:
                        if c.name not in count_by_prop:
                            count_by_prop[c.name] = c.count
                        else:
                            count_by_prop[c.name] += c.count

                add_counts(self.counts)
                add_counts(other.counts)
                new_counts = [LabelCount.create(k, v) for k, v in count_by_prop.items()]
        return LabelOverview.create(self.property_key, counts=new_counts)

    def to_dict(self) -> dict[str, Any]:
        """Returns this label overview as a dictionary."""
        return self.properties

    def __eq__(self, o: object) -> bool:
        if isinstance(o, LabelOverview):
            o = o.to_dict()

        if not isinstance(o, dict):
            return NotImplemented

        return self.to_dict() == o


class LabelExtension(ExtensionManagementMixin[pystac.Item | pystac.Collection]):
    """A class that can be used to extend the properties of an
    :class:`~pystac.Item` with properties from the :stac-ext:`Label Extension <label>`.

    To create an instance of :class:`LabeExtension`, use the
    :meth:`LabelExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> label_ext = LabelExtension.ext(item)
    """

    name: Literal["label"] = "label"
    obj: pystac.Item
    schema_uri: str

    def __init__(self, item: pystac.Item) -> None:
        self.obj = item
        self.schema_uri = SCHEMA_URI

    def apply(
        self,
        label_description: str,
        label_type: LabelType,
        label_properties: list[str] | None = None,
        label_classes: list[LabelClasses] | None = None,
        label_tasks: list[LabelTask | str] | None = None,
        label_methods: list[LabelMethod | str] | None = None,
        label_overviews: list[LabelOverview] | None = None,
    ) -> None:
        """Applies label extension properties to the extended Item.

        Args:
            label_description : A description of the label, how it was created,
                and what it is recommended for
            label_type : An Enum of either vector label type or raster label type. Use
                one of :class:`~pystac.LabelType`.
            label_properties : These are the names of the property field(s) in each
                Feature of the label asset's FeatureCollection that contains the classes
                (keywords from label:classes if the property defines classes).
                If labels are rasters, this should be None.
            label_classes : Optional, but required if using categorical data.
                A list of :class:`LabelClasses` instances defining the list of possible
                class names for each label:properties. (e.g., tree, building, car,
                hippo)
            label_tasks : Recommended to be a subset of 'regression', 'classification',
                'detection', or 'segmentation', but may be an arbitrary value.
            label_methods: Recommended to be a subset of 'automated' or 'manual',
                but may be an arbitrary value.
            label_overviews : Optional list of :class:`LabelOverview` instances
                that store counts (for classification-type data) or summary statistics
                (for continuous numerical/regression data).
        """
        self.label_description = label_description
        self.label_type = label_type
        self.label_properties = label_properties
        self.label_classes = label_classes
        self.label_tasks = label_tasks
        self.label_methods = label_methods
        self.label_overviews = label_overviews

    @property
    def label_description(self) -> str:
        """Gets or sets a description of the label, how it was created,
        and what it is recommended for."""
        return cast(
            str,
            get_required(
                self.obj.properties.get(DESCRIPTION_PROP), self.obj, DESCRIPTION_PROP
            ),
        )

    @label_description.setter
    def label_description(self, v: str) -> None:
        self.obj.properties[DESCRIPTION_PROP] = v

    @property
    def label_type(self) -> LabelType:
        """Gets or sets an Enum of either vector label type or raster label type."""
        return LabelType(
            get_required(self.obj.properties.get(TYPE_PROP), self.obj, TYPE_PROP)
        )

    @label_type.setter
    def label_type(self, v: LabelType) -> None:
        self.obj.properties[TYPE_PROP] = v

    @property
    def label_properties(self) -> list[str] | None:
        """Gets or sets the names of the property field(s) in each
        Feature of the label asset's FeatureCollection that contains the classes
        (keywords from label:classes if the property defines classes).
        If labels are rasters, this should be None."""
        return self.obj.properties.get(PROPERTIES_PROP)

    @label_properties.setter
    def label_properties(self, v: list[str] | None) -> None:
        self.obj.properties[PROPERTIES_PROP] = v

    @property
    def label_classes(self) -> list[LabelClasses] | None:
        """Gets or set a list of :class:`LabelClasses` defining the list of possible
        class names for each label:properties. (e.g., tree, building, car, hippo).

        Optional, but required if using categorical data."""
        label_classes = self.obj.properties.get(CLASSES_PROP)
        if label_classes is not None:
            return [LabelClasses(classes) for classes in label_classes]
        else:
            return None

    @label_classes.setter
    def label_classes(self, v: list[LabelClasses] | None) -> None:
        if v is None:
            self.obj.properties.pop(CLASSES_PROP, None)
        else:
            if not isinstance(v, list):
                raise pystac.STACError(
                    f"label_classes must be a list! Invalid input: {v}"
                )

            classes = [x.to_dict() for x in v]
            self.obj.properties[CLASSES_PROP] = classes

    @property
    def label_tasks(self) -> list[LabelTask | str] | None:
        """Gets or set a list of tasks these labels apply to. Usually a subset of
        'regression', 'classification', 'detection', or 'segmentation', but may be
        arbitrary values."""
        return self.obj.properties.get(TASKS_PROP)

    @label_tasks.setter
    def label_tasks(self, v: list[LabelTask | str] | None) -> None:
        if v is None:
            self.obj.properties.pop(TASKS_PROP, None)
        else:
            self.obj.properties[TASKS_PROP] = v

    @property
    def label_methods(self) -> list[LabelMethod | str] | None:
        """Gets or set a list of methods used for labeling.

        Usually a subset of 'automated' or 'manual', but may be arbitrary values."""
        return self.obj.properties.get("label:methods")

    @label_methods.setter
    def label_methods(self, v: list[LabelMethod | str] | None) -> None:
        if v is None:
            self.obj.properties.pop("label:methods", None)
        else:
            self.obj.properties["label:methods"] = v

    @property
    def label_overviews(self) -> list[LabelOverview] | None:
        """Gets or set a list of :class:`LabelOverview` instances
        that store counts (for classification-type data) or summary statistics (for
        continuous numerical/regression data)."""
        overviews = self.obj.properties.get(OVERVIEWS_PROP)
        if overviews is not None:
            return [LabelOverview(overview) for overview in overviews]
        else:
            return None

    @label_overviews.setter
    def label_overviews(self, v: list[LabelOverview] | None) -> None:
        if v is None:
            self.obj.properties.pop(OVERVIEWS_PROP, None)
        else:
            self.obj.properties[OVERVIEWS_PROP] = [x.to_dict() for x in v]

    def __repr__(self) -> str:
        return f"<LabelItemExt Item id={self.obj.id}>"

    def add_source(
        self,
        source_item: pystac.Item,
        title: str | None = None,
        assets: list[str] | None = None,
    ) -> None:
        """Adds a link to a source item.

        Args:
            source_item : Source imagery that the LabelItem applies to.
            title : Optional title for the link.
            assets : Optional list of assets that determine what
                assets in the source item this label item data applies to.
        """
        extra_fields = None
        if assets is not None:
            extra_fields = {"label:assets": assets}
        link = pystac.Link(
            "source",
            source_item,
            title=title,
            media_type=pystac.MediaType.JSON,
            extra_fields=extra_fields,
        )
        self.obj.add_link(link)

    def get_sources(self) -> Iterable[pystac.Item]:
        """Gets any source items that describe the source imagery used to generate
        this LabelItem.

        Returns:
            A possibly empty list of source imagery items. Determined by links of this
            LabelItem that have ``rel=='source'``.
        """
        return map(lambda x: cast(pystac.Item, x), self.obj.get_stac_objects("source"))

    def add_labels(
        self,
        href: str,
        title: str | None = None,
        media_type: str | None = None,
        properties: dict[str, Any] | None = None,
    ) -> None:
        """Adds a label asset to this LabelItem.

        Args:
            href : Link to the asset object. Relative and absolute links are both
                allowed.
            title : Optional displayed title for clients and users.
            media_type : Optional description of the media type. Registered Media
                Types are preferred. See :class:`~pystac.MediaType` for common
                media types.
            properties : Optional, additional properties for this asset. This is
                used by extensions as a way to serialize and deserialize properties on
                asset object JSON.
        """

        self.obj.add_asset(
            "labels",
            pystac.Asset(
                href=href, title=title, media_type=media_type, extra_fields=properties
            ),
        )

    def add_geojson_labels(
        self,
        href: str,
        title: str | None = None,
        properties: dict[str, Any] | None = None,
    ) -> None:
        """Adds a GeoJSON label asset to this LabelItem.

        Args:
            href : Link to the asset object. Relative and absolute links are both
                allowed.
            title : Optional displayed title for clients and users.
            properties : Optional, additional properties for this asset. This is
                used by extensions as a way to serialize and deserialize properties on
                asset object JSON.
        """
        self.add_labels(
            href,
            title=title,
            properties=properties,
            media_type=pystac.MediaType.GEOJSON,
        )

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def get_schema_uris(cls) -> list[str]:
        warnings.warn(
            "get_schema_uris is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return SCHEMA_URIS

    @classmethod
    def ext(cls, obj: pystac.Item, add_if_missing: bool = False) -> LabelExtension:
        """Extends the given STAC Object with properties from the :stac-ext:`Label
        Extension <label>`.

        This extension can be applied to instances of :class:`~pystac.Item`.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cls(obj)
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesLabelExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesLabelExtension(obj)


class SummariesLabelExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Label Extension <label>`.
    """

    @property
    def label_properties(self) -> list[str] | None:
        """Get or sets the summary of :attr:`LabelExtension.label_properties` values
        for this Collection.
        """

        return self.summaries.get_list(PROPERTIES_PROP)

    @label_properties.setter
    def label_properties(self, v: list[str] | None) -> None:
        self._set_summary(PROPERTIES_PROP, v)

    @property
    def label_classes(self) -> list[LabelClasses] | None:
        """Get or sets the summary of :attr:`LabelExtension.label_classes` values
        for this Collection.
        """

        return map_opt(
            lambda classes: [LabelClasses(c) for c in classes],
            self.summaries.get_list(CLASSES_PROP),
        )

    @label_classes.setter
    def label_classes(self, v: list[LabelClasses] | None) -> None:
        self._set_summary(
            CLASSES_PROP, map_opt(lambda classes: [c.to_dict() for c in classes], v)
        )

    @property
    def label_type(self) -> list[LabelType] | None:
        """Get or sets the summary of :attr:`LabelExtension.label_type` values
        for this Collection.
        """

        return self.summaries.get_list(TYPE_PROP)

    @label_type.setter
    def label_type(self, v: list[LabelType] | None) -> None:
        self._set_summary(TYPE_PROP, v)

    @property
    def label_tasks(self) -> list[LabelTask | str] | None:
        """Get or sets the summary of :attr:`LabelExtension.label_tasks` values
        for this Collection.
        """

        return self.summaries.get_list(TASKS_PROP)

    @label_tasks.setter
    def label_tasks(self, v: list[LabelTask | str] | None) -> None:
        self._set_summary(TASKS_PROP, v)

    @property
    def label_methods(self) -> list[LabelMethod | str] | None:
        """Get or sets the summary of :attr:`LabelExtension.label_methods` values
        for this Collection.
        """

        return self.summaries.get_list(METHODS_PROP)

    @label_methods.setter
    def label_methods(self, v: list[LabelMethod | str] | None) -> None:
        self._set_summary(METHODS_PROP, v)


class LabelExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {
        "label",
        *[uri for uri in SCHEMA_URIS if uri != SCHEMA_URI],
    }
    stac_object_types = {pystac.STACObjectType.ITEM}

    def get_object_links(
        self, obj: pystac.STACObject
    ) -> list[str | pystac.RelType] | None:
        if isinstance(obj, pystac.Item):
            return [LabelRelType.SOURCE]
        return None

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        if info.object_type == pystac.STACObjectType.ITEM and version < "1.0.0":
            props = obj["properties"]
            # Migrate 0.8.0-rc1 non-pluralized forms
            # As it's a common mistake, convert for any pre-1.0.0 version.
            if "label:property" in props and PROPERTIES_PROP not in props:
                props[PROPERTIES_PROP] = props["label:property"]
                del props["label:property"]

            if "label:task" in props and TASKS_PROP not in props:
                props[TASKS_PROP] = props["label:task"]
                del props["label:task"]

            if "label:overview" in props and OVERVIEWS_PROP not in props:
                props[OVERVIEWS_PROP] = props["label:overview"]
                del props["label:overview"]

            if "label:method" in props and "label:methods" not in props:
                props["label:methods"] = props["label:method"]
                del props["label:method"]

        super().migrate(obj, version, info)


LABEL_EXTENSION_HOOKS: ExtensionHooks = LabelExtensionHooks()



from typing import Any

from pystac.utils import StringEnum


class ProviderRole(StringEnum):
    """Enumerates the allows values of the Provider "role" field."""

    LICENSOR = "licensor"
    PRODUCER = "producer"
    PROCESSOR = "processor"
    HOST = "host"


class Provider:
    """Provides information about a provider of STAC data. A provider is any of the
    organizations that captured or processed the content of the collection and therefore
    influenced the data offered by this collection. May also include information about
    the final storage provider hosting the data.

    Args:
        name : The name of the organization or the individual.
        description : Optional multi-line description to add further provider
            information such as processing details for processors and producers,
            hosting details for hosts or basic contact information.
        roles : Optional roles of the provider. Any of
            licensor, producer, processor or host.
        url : Optional homepage on which the provider describes the dataset
            and publishes contact information.
        extra_fields : Optional dictionary containing additional top-level fields
            defined on the Provider object.
    """

    name: str
    """The name of the organization or the individual."""

    description: str | None
    """Optional multi-line description to add further provider
    information such as processing details for processors and producers,
    hosting details for hosts or basic contact information."""

    roles: list[ProviderRole] | None
    """Optional roles of the provider. Any of
    licensor, producer, processor or host."""

    url: str | None
    """Optional homepage on which the provider describes the dataset
    and publishes contact information."""

    extra_fields: dict[str, Any]
    """Dictionary containing additional top-level fields defined on the Provider
    object."""

    def __init__(
        self,
        name: str,
        description: str | None = None,
        roles: list[ProviderRole] | None = None,
        url: str | None = None,
        extra_fields: dict[str, Any] | None = None,
    ):
        self.name = name
        self.description = description
        self.roles = roles
        self.url = url
        self.extra_fields = extra_fields or {}

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Provider):
            return NotImplemented
        return self.to_dict() == o.to_dict()

    def _repr_html_(self) -> str:
        from html import escape

        from pystac.html.jinja_env import get_jinja_env

        jinja_env = get_jinja_env()
        if jinja_env:
            template = jinja_env.get_template("JSON.jinja2")
            return str(template.render(dict=self.to_dict(), plain=escape(repr(self))))
        else:
            return escape(repr(self))

    def to_dict(self) -> dict[str, Any]:
        """Returns this provider as a dictionary.

        Returns:
            dict: A serialization of the Provider.
        """
        d: dict[str, Any] = {"name": self.name}
        if self.description is not None:
            d["description"] = self.description
        if self.roles is not None:
            d["roles"] = self.roles
        if self.url is not None:
            d["url"] = self.url

        d.update(self.extra_fields)

        return d

    @staticmethod
    def from_dict(d: dict[str, Any]) -> "Provider":
        """Constructs a Provider from a dict.

        Returns:
            Provider: The Provider deserialized from the JSON dict.
        """
        return Provider(
            name=d["name"],
            description=d.get("description"),
            roles=d.get(
                "roles",
            ),
            url=d.get("url"),
            extra_fields={
                k: v
                for k, v in d.items()
                if k not in {"name", "description", "roles", "url"}
            },
        )



import json
from datetime import datetime

import pytest

import pystac
from pystac import Collection, ExtensionTypeError, Item
from pystac.extensions.timestamps import TimestampsExtension
from pystac.summaries import RangeSummary
from pystac.utils import datetime_to_str, get_opt, str_to_datetime
from tests.utils import TestCases, assert_to_from_dict

EXAMPLE_URI = TestCases.get_path("data-files/timestamps/example-landsat8.json")
SAMPLE_DATETIME_STR = "2020-01-01T00:00:00Z"


@pytest.fixture
def item() -> Item:
    return Item.from_file(EXAMPLE_URI)


@pytest.fixture
def sample_datetime() -> datetime:
    return str_to_datetime(SAMPLE_DATETIME_STR)


def test_to_from_dict() -> None:
    with open(EXAMPLE_URI) as f:
        item_dict = json.load(f)
    assert_to_from_dict(pystac.Item, item_dict)


def test_apply() -> None:
    item = next(TestCases.case_2().get_items(recursive=True))
    assert not TimestampsExtension.has_extension(item)

    TimestampsExtension.add_to(item)
    assert TimestampsExtension.has_extension(item)
    TimestampsExtension.ext(item).apply(
        published=str_to_datetime("2020-01-03T06:45:55Z"),
        expires=str_to_datetime("2020-02-03T06:45:55Z"),
        unpublished=str_to_datetime("2020-03-03T06:45:55Z"),
    )

    for d in [
        TimestampsExtension.ext(item).published,
        TimestampsExtension.ext(item).expires,
        TimestampsExtension.ext(item).unpublished,
    ]:
        assert isinstance(d, datetime)

    for p in ("published", "expires", "unpublished"):
        assert isinstance(item.properties[p], str)

    published_str = "2020-04-03T06:45:55Z"
    TimestampsExtension.ext(item).apply(published=str_to_datetime(published_str))
    assert isinstance(TimestampsExtension.ext(item).published, datetime)
    assert item.properties["published"] == published_str

    for d in [
        TimestampsExtension.ext(item).expires,
        TimestampsExtension.ext(item).unpublished,
    ]:
        assert d is None

    for p in ("expires", "unpublished"):
        assert p not in item.properties


@pytest.mark.vcr()
def test_validate_timestamps(item: Item) -> None:
    item.validate()


@pytest.mark.vcr()
def test_expires(item: Item, sample_datetime: datetime) -> None:
    # Get
    assert "expires" in item.properties
    timestamps_expires = TimestampsExtension.ext(item).expires
    assert isinstance(timestamps_expires, datetime)
    assert datetime_to_str(get_opt(timestamps_expires)) == item.properties["expires"]

    # Set
    TimestampsExtension.ext(item).expires = sample_datetime
    assert SAMPLE_DATETIME_STR == item.properties["expires"]

    # Get from Asset
    asset_no_prop = item.assets["red"]
    asset_prop = item.assets["blue"]
    assert (
        TimestampsExtension.ext(asset_no_prop).expires
        == TimestampsExtension.ext(item).expires
    )
    assert TimestampsExtension.ext(asset_prop).expires == str_to_datetime(
        "2018-12-02T00:00:00Z"
    )

    # # Set to Asset
    asset_value = str_to_datetime("2019-02-02T00:00:00Z")
    TimestampsExtension.ext(asset_no_prop).expires = asset_value
    assert (
        TimestampsExtension.ext(asset_no_prop).expires
        != TimestampsExtension.ext(item).expires
    )
    assert TimestampsExtension.ext(asset_no_prop).expires == asset_value

    # Validate
    item.validate()


@pytest.mark.vcr()
def test_published(item: Item, sample_datetime: datetime) -> None:
    # Get
    assert "published" in item.properties
    timestamps_published = TimestampsExtension.ext(item).published
    assert isinstance(timestamps_published, datetime)
    assert (
        datetime_to_str(get_opt(timestamps_published)) == item.properties["published"]
    )

    # Set
    TimestampsExtension.ext(item).published = sample_datetime
    assert SAMPLE_DATETIME_STR == item.properties["published"]

    # Get from Asset
    asset_no_prop = item.assets["red"]
    asset_prop = item.assets["blue"]
    assert (
        TimestampsExtension.ext(asset_no_prop).published
        == TimestampsExtension.ext(item).published
    )
    assert TimestampsExtension.ext(asset_prop).published == str_to_datetime(
        "2018-11-02T00:00:00Z"
    )

    # # Set to Asset
    asset_value = str_to_datetime("2019-02-02T00:00:00Z")
    TimestampsExtension.ext(asset_no_prop).published = asset_value
    assert (
        TimestampsExtension.ext(asset_no_prop).published
        != TimestampsExtension.ext(item).published
    )
    assert TimestampsExtension.ext(asset_no_prop).published == asset_value

    # Validate
    item.validate()


@pytest.mark.vcr()
def test_unpublished(item: Item, sample_datetime: datetime) -> None:
    # Get
    assert "unpublished" not in item.properties
    timestamps_unpublished = TimestampsExtension.ext(item).unpublished
    assert timestamps_unpublished is None

    # Set
    TimestampsExtension.ext(item).unpublished = sample_datetime
    assert SAMPLE_DATETIME_STR == item.properties["unpublished"]

    # Get from Asset
    asset_no_prop = item.assets["red"]
    asset_prop = item.assets["blue"]
    assert (
        TimestampsExtension.ext(asset_no_prop).unpublished
        == TimestampsExtension.ext(item).unpublished
    )
    assert TimestampsExtension.ext(asset_prop).unpublished == str_to_datetime(
        "2019-01-02T00:00:00Z"
    )

    # Set to Asset
    asset_value = str_to_datetime("2019-02-02T00:00:00Z")
    TimestampsExtension.ext(asset_no_prop).unpublished = asset_value
    assert (
        TimestampsExtension.ext(asset_no_prop).unpublished
        != TimestampsExtension.ext(item).unpublished
    )
    assert TimestampsExtension.ext(asset_no_prop).unpublished == asset_value

    # Validate
    item.validate()


def test_extension_not_implemented(item: Item) -> None:
    # Should raise exception if Item does not include extension URI
    item.stac_extensions.remove(TimestampsExtension.get_schema_uri())

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = TimestampsExtension.ext(item)

    # Should raise exception if owning Item does not include extension URI
    asset = item.assets["blue"]

    with pytest.raises(pystac.ExtensionNotImplemented):
        _ = TimestampsExtension.ext(asset)

    # Should succeed if Asset has no owner
    ownerless_asset = pystac.Asset.from_dict(asset.to_dict())
    _ = TimestampsExtension.ext(ownerless_asset)


def test_item_ext_add_to(item: Item) -> None:
    item.stac_extensions.remove(TimestampsExtension.get_schema_uri())
    assert TimestampsExtension.get_schema_uri() not in item.stac_extensions

    _ = TimestampsExtension.ext(item, add_if_missing=True)

    assert TimestampsExtension.get_schema_uri() in item.stac_extensions


def test_asset_ext_add_to(item: Item) -> None:
    item.stac_extensions.remove(TimestampsExtension.get_schema_uri())
    assert TimestampsExtension.get_schema_uri() not in item.stac_extensions
    asset = item.assets["blue"]

    _ = TimestampsExtension.ext(asset, add_if_missing=True)

    assert TimestampsExtension.get_schema_uri() in item.stac_extensions


def test_should_raise_exception_when_passing_invalid_extension_object() -> None:
    with pytest.raises(
        ExtensionTypeError,
        match=r"^TimestampsExtension does not apply to type 'object'$",
    ):
        # calling it wrong on purpose --------------v
        TimestampsExtension.ext(object())  # type: ignore


def test_item_repr(item: Item) -> None:
    assert (
        TimestampsExtension.ext(item).__repr__()
        == f"<ItemTimestampsExtension Item id={item.id}>"
    )


def test_asset_repr(item: Item) -> None:
    asset = item.assets["blue"]
    assert (
        TimestampsExtension.ext(asset).__repr__()
        == f"<AssetTimestampsExtension Asset href={asset.href}>"
    )


def test_summaries_published(multi_extent_collection: Collection) -> None:
    summaries_ext = TimestampsExtension.summaries(multi_extent_collection, True)
    published_range = RangeSummary(
        str_to_datetime("2020-01-01T00:00:00.000Z"),
        str_to_datetime("2020-01-02T00:00:00.000Z"),
    )

    summaries_ext.published = published_range

    assert summaries_ext.published == published_range

    summaries_dict = multi_extent_collection.to_dict()["summaries"]

    assert summaries_dict["published"] == {
        "minimum": datetime_to_str(published_range.minimum),
        "maximum": datetime_to_str(published_range.maximum),
    }


def test_summaries_expires(multi_extent_collection: Collection) -> None:
    summaries_ext = TimestampsExtension.summaries(multi_extent_collection, True)
    expires_range = RangeSummary(
        str_to_datetime("2020-01-01T00:00:00.000Z"),
        str_to_datetime("2020-01-02T00:00:00.000Z"),
    )

    summaries_ext.expires = expires_range

    assert summaries_ext.expires == expires_range

    summaries_dict = multi_extent_collection.to_dict()["summaries"]

    assert summaries_dict["expires"] == {
        "minimum": datetime_to_str(expires_range.minimum),
        "maximum": datetime_to_str(expires_range.maximum),
    }


def test_summaries_unpublished(multi_extent_collection: Collection) -> None:
    summaries_ext = TimestampsExtension.summaries(multi_extent_collection, True)
    unpublished_range = RangeSummary(
        str_to_datetime("2020-01-01T00:00:00.000Z"),
        str_to_datetime("2020-01-02T00:00:00.000Z"),
    )

    summaries_ext.unpublished = unpublished_range

    assert summaries_ext.unpublished == unpublished_range

    summaries_dict = multi_extent_collection.to_dict()["summaries"]

    assert summaries_dict["unpublished"] == {
        "minimum": datetime_to_str(unpublished_range.minimum),
        "maximum": datetime_to_str(unpublished_range.maximum),
    }


def test_summaries_adds_uri(multi_extent_collection: Collection) -> None:
    multi_extent_collection.stac_extensions = []
    with pytest.raises(
        pystac.ExtensionNotImplemented,
        match="Extension 'timestamps' is not implemented",
    ):
        TimestampsExtension.summaries(multi_extent_collection, add_if_missing=False)

    _ = TimestampsExtension.summaries(multi_extent_collection, True)

    assert (
        TimestampsExtension.get_schema_uri() in multi_extent_collection.stac_extensions
    )

    TimestampsExtension.remove_from(multi_extent_collection)
    assert (
        TimestampsExtension.get_schema_uri()
        not in multi_extent_collection.stac_extensions
    )


@pytest.mark.parametrize(
    "schema_uri",
    ("https://stac-extensions.github.io/timestamps/v1.0.0/schema.json",),
)
def test_migrate(schema_uri: str, item: Item) -> None:
    item_dict = item.to_dict(include_self_link=False, transform_hrefs=False)
    item_dict["stac_extensions"] = [schema_uri]
    item = Item.from_dict(item_dict, migrate=True)
    assert item.stac_extensions == [
        "https://stac-extensions.github.io/timestamps/v1.1.0/schema.json"
    ]



from __future__ import annotations

import copy
import json
import os
import pickle
import tempfile
from copy import deepcopy
from pathlib import Path
from typing import Any, cast

import dateutil.relativedelta
import pytest

import pystac
import pystac.serialization.common_properties
from pystac import Asset, Catalog, Collection, Item, Link, STACValidationError
from pystac.utils import (
    datetime_to_str,
    get_opt,
    is_absolute_href,
    make_posix_style,
    str_to_datetime,
)
from pystac.validation import validate_dict
from tests.utils import TestCases, assert_to_from_dict


def test_to_from_dict(sample_item_dict: dict[str, Any]) -> None:
    param_dict = deepcopy(sample_item_dict)

    assert_to_from_dict(Item, param_dict)
    item = Item.from_dict(param_dict)
    assert item.id == "CS3-20160503_132131_05"

    # test asset creation additional field(s)
    assert (
        item.assets["analytic"].extra_fields["product"]
        == "http://cool-sat.com/catalog/products/analytic.json"
    )
    assert len(item.assets["thumbnail"].extra_fields) == 0

    # test that the parameter is preserved
    assert param_dict == sample_item_dict

    # assert that the parameter is preserved regardless of preserve_dict
    Item.from_dict(param_dict, preserve_dict=False)
    assert param_dict == sample_item_dict


def test_from_dict_set_root(sample_item_dict: dict[str, Any]) -> None:
    catalog = pystac.Catalog(id="test", description="test desc")
    item = Item.from_dict(sample_item_dict, root=catalog)
    assert item.get_root() is catalog


def test_set_self_href_does_not_break_asset_hrefs() -> None:
    cat = TestCases.case_2()
    for item in cat.get_items(recursive=True):
        for asset in item.assets.values():
            if is_absolute_href(asset.href):
                asset.href = f"./{os.path.basename(asset.href)}"
        item.set_self_href("http://example.com/item.json")
        for asset in item.assets.values():
            assert is_absolute_href(asset.href)


def test_set_self_href_none_ignores_relative_asset_hrefs() -> None:
    cat = TestCases.case_2()
    for item in cat.get_items(recursive=True):
        for asset in item.assets.values():
            if is_absolute_href(asset.href):
                asset.href = f"./{os.path.basename(asset.href)}"
        item.set_self_href(None)
        for asset in item.assets.values():
            assert not is_absolute_href(asset.href)


def test_asset_absolute_href(sample_item: Item) -> None:
    item_path = TestCases.get_path("data-files/item/sample-item.json")
    sample_item.set_self_href(item_path)
    rel_asset = Asset("./data.geojson")
    rel_asset.set_owner(sample_item)
    expected_href = make_posix_style(
        os.path.abspath(os.path.join(os.path.dirname(item_path), "./data.geojson"))
    )
    actual_href = rel_asset.get_absolute_href()
    assert expected_href == actual_href


def test_asset_absolute_href_no_item_self(sample_item_dict: dict[str, Any]) -> None:
    item = Item.from_dict(sample_item_dict)
    assert item.get_self_href() is None

    rel_asset = Asset("./data.geojson")
    rel_asset.set_owner(item)
    actual_href = rel_asset.get_absolute_href()
    assert actual_href is None


def test_item_field_order() -> None:
    item = pystac.Item.from_file(TestCases.get_path("data-files/item/sample-item.json"))
    item_dict = item.to_dict(include_self_link=False)
    expected_order = [
        "type",
        "stac_version",
        "stac_extensions",
        "id",
        "geometry",
        "bbox",
        "properties",
        "links",
        "assets",
        "collection",
    ]
    actual_order = list(item_dict.keys())
    assert actual_order == expected_order


def test_extra_fields() -> None:
    item = pystac.Item.from_file(TestCases.get_path("data-files/item/sample-item.json"))

    item.extra_fields["test"] = "extra"

    with tempfile.TemporaryDirectory() as tmp_dir:
        p = os.path.join(tmp_dir, "item.json")
        item.save_object(include_self_link=False, dest_href=p)
        with open(p) as f:
            item_json = json.load(f)
        assert "test" in item_json
        assert item_json["test"] == "extra"

        read_item = pystac.Item.from_file(p)
        assert "test" in read_item.extra_fields
        assert read_item.extra_fields["test"] == "extra"


def test_clearing_collection() -> None:
    collection = TestCases.case_4().get_child("acc")
    assert isinstance(collection, pystac.Collection)
    item = next(collection.get_items(recursive=True))
    assert item.collection_id == collection.id
    item.set_collection(None)
    assert item.collection_id is None
    assert item.get_collection() is None
    item.set_collection(collection)
    assert item.collection_id == collection.id
    assert item.get_collection() is collection


def test_datetime_ISO8601_format(sample_item: Item) -> None:
    formatted_time = sample_item.to_dict()["properties"]["datetime"]
    assert "2016-05-03T13:22:30.040000Z" == formatted_time


@pytest.mark.vcr()
def test_null_datetime() -> None:
    item = pystac.Item.from_file(TestCases.get_path("data-files/item/sample-item.json"))

    with pytest.raises(pystac.STACError):
        Item(
            "test",
            geometry=item.geometry,
            bbox=item.bbox,
            datetime=None,
            properties={},
        )

    null_dt_item = Item(
        "test",
        geometry=item.geometry,
        bbox=item.bbox,
        datetime=None,
        properties={
            "start_datetime": datetime_to_str(get_opt(item.datetime)),
            "end_datetime": datetime_to_str(get_opt(item.datetime)),
        },
    )

    null_dt_item.validate()


def test_get_assets() -> None:
    item = pystac.Item.from_file(TestCases.get_path("data-files/item/sample-item.json"))

    media_type_filter = item.get_assets(media_type=pystac.MediaType.COG)
    assert set(media_type_filter.keys()) == {"analytic"}
    role_filter = item.get_assets(role="data")
    assert set(role_filter.keys()) == {"analytic"}
    multi_filter = item.get_assets(media_type=pystac.MediaType.PNG, role="thumbnail")
    assert set(multi_filter.keys()) == {"thumbnail"}
    multi_filter["thumbnail"].description = "foo"
    assert item.assets["thumbnail"].description != "foo"

    no_filter = item.get_assets()
    assert set(no_filter.keys()) == {"analytic", "thumbnail"}
    no_assets = item.get_assets(media_type=pystac.MediaType.HDF)
    assert no_assets == {}


@pytest.mark.vcr()
def test_null_datetime_constructor() -> None:
    item = pystac.Item.from_file(TestCases.get_path("data-files/item/sample-item.json"))
    with pytest.raises(pystac.STACError):
        Item(
            "test",
            geometry=item.geometry,
            bbox=item.bbox,
            datetime=None,
            end_datetime=item.datetime,
            properties={},
        )
    with pytest.raises(pystac.STACError):
        Item(
            "test",
            geometry=item.geometry,
            bbox=item.bbox,
            datetime=None,
            start_datetime=item.datetime,
            properties={},
        )
    assert item.datetime
    null_dt_item = Item(
        "test",
        geometry=item.geometry,
        bbox=item.bbox,
        datetime=None,
        start_datetime=item.datetime,
        end_datetime=item.datetime + dateutil.relativedelta.relativedelta(days=1),
        properties={},
    )
    null_dt_item.validate()


def test_get_set_asset_datetime() -> None:
    item = pystac.Item.from_file(
        TestCases.get_path("data-files/item/sample-item-asset-properties.json")
    )
    item_datetime = item.datetime

    # No property on asset
    assert item.get_datetime(item.assets["thumbnail"]) == item.datetime

    # Property on asset
    assert item.get_datetime(item.assets["analytic"]) != item.datetime
    assert item.get_datetime(item.assets["analytic"]) == str_to_datetime(
        "2017-05-03T13:22:30.040Z"
    )

    item.set_datetime(
        str_to_datetime("2018-05-03T13:22:30.040Z"), item.assets["thumbnail"]
    )
    assert item.get_datetime() == item_datetime
    assert item.get_datetime(item.assets["thumbnail"]) == str_to_datetime(
        "2018-05-03T13:22:30.040Z"
    )


def test_read_eo_item_owns_asset() -> None:
    item = next(TestCases.case_1().get_items(recursive=True))
    assert len(item.assets) > 0
    for asset_key in item.assets:
        assert item.assets[asset_key].owner == item


@pytest.mark.vcr()
def test_null_geometry() -> None:
    m = TestCases.get_path(
        "data-files/examples/1.0.0-beta.2/item-spec/examples/null-geom-item.json"
    )
    with open(m) as f:
        item_dict = json.load(f)

    validate_dict(item_dict, pystac.STACObjectType.ITEM)

    item = Item.from_dict(item_dict)
    assert isinstance(item, Item)
    item.validate()

    item_dict = item.to_dict()
    assert item_dict["geometry"] is None
    assert "bbox" not in item_dict


def test_0_9_item_with_no_extensions_does_not_read_collection_data() -> None:
    item_json = pystac.StacIO.default().read_json(
        TestCases.get_path("data-files/examples/hand-0.9.0/010100/010100.json")
    )
    assert item_json.get("stac_extensions") is None
    assert item_json.get("stac_version") == "0.9.0"

    did_merge = pystac.serialization.common_properties.merge_common_properties(
        item_json
    )
    assert not did_merge


def test_clone_preserves_assets() -> None:
    cat = TestCases.case_2()
    original_item = next(cat.get_items(recursive=True))
    assert len(original_item.assets) > 0
    assert all(asset.owner is original_item for asset in original_item.assets.values())

    cloned_item = original_item.clone()

    for key in original_item.assets:
        assert key in cloned_item.assets, f"Failed to preserve asset {key}"
        cloned_asset = cloned_item.assets.get(key)
        if cloned_asset is not None:
            assert cloned_asset.owner is cloned_item, f"Failed set owner for {key}"


def test_make_asset_href_relative_is_noop_on_relative_hrefs() -> None:
    cat = TestCases.case_2()
    item = next(cat.get_items(recursive=True))
    asset = list(item.assets.values())[0]
    assert not is_absolute_href(asset.href)
    original_href = asset.get_absolute_href()

    item.make_asset_hrefs_relative()
    assert asset.get_absolute_href() == original_href


def test_from_invalid_dict_raises_exception() -> None:
    stac_io = pystac.StacIO.default()
    catalog_dict = stac_io.read_json(
        TestCases.get_path("data-files/catalogs/test-case-1/catalog.json")
    )
    with pytest.raises(pystac.STACTypeError):
        _ = pystac.Item.from_dict(catalog_dict)


@pytest.mark.vcr()
def test_relative_extension_path() -> None:
    item = pystac.Item.from_file(
        TestCases.get_path(
            "data-files/item/sample-item-with-relative-extension-path.json"
        )
    )
    item.validate()


class TestItemSubClass:
    """This tests cases related to creating classes inheriting from pystac.Catalog to
    ensure that inheritance, class methods, etc. function as expected."""

    SAMPLE_ITEM = TestCases.get_path("data-files/item/sample-item.json")

    class BasicCustomItem(pystac.Item):
        pass

    def test_from_dict_returns_subclass(self) -> None:
        stac_io = pystac.StacIO.default()
        item_dict = stac_io.read_json(self.SAMPLE_ITEM)
        custom_item = self.BasicCustomItem.from_dict(item_dict)

        assert isinstance(custom_item, self.BasicCustomItem)

    def test_from_file_returns_subclass(self) -> None:
        custom_item = self.BasicCustomItem.from_file(self.SAMPLE_ITEM)

        assert isinstance(custom_item, self.BasicCustomItem)

    def test_clone(self) -> None:
        custom_item = self.BasicCustomItem.from_file(self.SAMPLE_ITEM)
        cloned_item = custom_item.clone()

        assert isinstance(cloned_item, self.BasicCustomItem)


def test_asset_clone() -> None:
    with open(TestCases.get_path("data-files/item/sample-item.json")) as src:
        item_dict = json.load(src)
    asset_dict = item_dict["assets"]["analytic"]
    original_asset = Asset.from_dict(asset_dict)

    cloned_asset = original_asset.clone()

    assert original_asset.to_dict() == asset_dict
    assert cloned_asset.to_dict() == asset_dict

    # Changes to original asset should not affect cloned Asset
    original_asset.description = "Some new description"
    original_asset.href = "/path/to/new/href"
    original_asset.title = "New Title"
    original_asset.roles = ["new role"]
    original_asset.roles.append("new role")
    original_asset.extra_fields["new_field"] = "new_value"
    assert cloned_asset.to_dict() == asset_dict


class TestAssetSubClass:
    class CustomAsset(Asset):
        pass

    AssetDict = dict[str, str | list[str]]

    @pytest.fixture
    def asset_dict(self) -> AssetDict:
        with open(TestCases.get_path("data-files/item/sample-item.json")) as src:
            item_dict = json.load(src)
        return cast(TestAssetSubClass.AssetDict, item_dict["assets"]["analytic"])

    def test_from_dict(self, asset_dict: AssetDict) -> None:
        asset = self.CustomAsset.from_dict(asset_dict)
        assert isinstance(asset, self.CustomAsset)

    def test_clone(self, asset_dict: AssetDict) -> None:
        asset = self.CustomAsset.from_dict(asset_dict)
        cloned_asset = asset.clone()
        assert isinstance(cloned_asset, self.CustomAsset)


def test_custom_item_from_dict(item: Item) -> None:
    # https://github.com/stac-utils/pystac/issues/862
    class CustomItem(Item):
        @classmethod
        def from_dict(
            cls,
            d: dict[str, Any],
            href: str | None = None,
            root: Catalog | None = None,
            migrate: bool = True,
            preserve_dict: bool = True,
        ) -> CustomItem:
            return super().from_dict(d)

    _ = CustomItem.from_dict(item.to_dict())


def test_item_from_dict_raises_useful_error() -> None:
    item_dict = {"type": "Feature", "stac_version": "1.1.0", "id": "lalalalala"}
    with pytest.raises(pystac.STACError, match="Invalid Item: "):
        Item.from_dict(item_dict)


def test_item_from_dict_with_missing_stac_version_raises_useful_error() -> None:
    item_dict = {"type": "Feature", "id": "lalalalala"}
    with pytest.raises(pystac.STACTypeError, match="'stac_version' is missing"):
        Item.from_dict(item_dict, migrate=False)


def test_item_from_dict_with_missing_type_raises_useful_error() -> None:
    item_dict = {"stac_version": "0.8.0", "id": "lalalalala"}
    with pytest.raises(pystac.STACTypeError, match="'type' is missing"):
        Item.from_dict(item_dict, migrate=False)


@pytest.mark.parametrize("add_canonical", (True, False))
def test_remove_hierarchical_links(
    test_case_1_catalog: Catalog, add_canonical: bool
) -> None:
    item = next(test_case_1_catalog.get_items(recursive=True))
    item.remove_hierarchical_links(add_canonical=add_canonical)
    for link in item.links:
        assert not link.is_hierarchical()
    assert bool(item.get_single_link("canonical")) == add_canonical


def test_geo_interface() -> None:
    item = pystac.Item.from_file(TestCases.get_path("data-files/item/sample-item.json"))
    assert (
        item.to_dict(include_self_link=False, transform_hrefs=False)
        == item.__geo_interface__
    )


def test_duplicate_self_links(tmp_path: Path, sample_item: pystac.Item) -> None:
    # https://github.com/stac-utils/pystac/issues/1102
    assert len(sample_item.get_links(rel="self")) == 1
    path = tmp_path / "item.json"
    sample_item.save_object(include_self_link=True, dest_href=str(path))
    sample_item = Item.from_file(path)
    assert len(sample_item.get_links(rel="self")) == 1


def test_get_derived_from_when_none_exists(test_case_1_catalog: Catalog) -> None:
    item = next(test_case_1_catalog.get_items(recursive=True))
    assert item.get_derived_from() == []
    for link in item.links:
        assert link.rel != pystac.RelType.DERIVED_FROM
    assert item.get_single_link(pystac.RelType.DERIVED_FROM) is None


def test_add_derived_from(test_case_1_catalog: Catalog) -> None:
    items = list(test_case_1_catalog.get_items(recursive=True))
    item_0 = items[0]
    item_1 = items[1]
    item_2 = items[2]
    item_0.add_derived_from(item_1, item_2.self_href)
    derived_from = item_0.get_derived_from()
    assert len(derived_from) == 2
    assert derived_from[0].id == item_1.id
    assert derived_from[1].id == item_2.id
    filtered = [
        link for link in item_0.links if link.rel == pystac.RelType.DERIVED_FROM
    ]
    assert len(filtered) == 2
    assert filtered[0].to_dict(transform_href=False)["href"] == item_1.self_href
    assert filtered[1].to_dict(transform_href=False)["href"] == item_2.self_href


def test_get_unresolvable_derived_from(test_case_1_catalog: Catalog) -> None:
    item = next(test_case_1_catalog.get_items(recursive=True))
    item.add_derived_from("foo")
    with pytest.raises(
        pystac.STACError, match="Link failed to resolve. Use get_links instead"
    ):
        item.get_derived_from()

    links = item.get_links(pystac.RelType.DERIVED_FROM)
    assert len(links) == 1


def test_remove_unresolvable_derived_from(test_case_1_catalog: Catalog) -> None:
    item = next(test_case_1_catalog.get_items(recursive=True))
    item.add_derived_from("foo")
    with pytest.raises(
        pystac.STACError, match="Link failed to resolve. Use remove_links instead"
    ):
        item.remove_derived_from("foo")

    item.remove_links(pystac.RelType.DERIVED_FROM)
    assert item.get_derived_from() == []


def test_remove_derived_from(test_case_1_catalog: Catalog) -> None:
    items = list(test_case_1_catalog.get_items(recursive=True))
    item_0 = items[0]
    item_1 = items[1]
    item_0.add_derived_from(item_1)
    item_0.remove_derived_from(item_1.id)
    assert item_0.get_derived_from() == []
    for link in item_0.links:
        assert link.rel != pystac.RelType.DERIVED_FROM
    assert item_0.get_single_link(pystac.RelType.DERIVED_FROM) is None


def test_delete_asset(tmp_asset: Asset) -> None:
    asset = tmp_asset
    href = asset.get_absolute_href()
    item = asset.owner

    assert href is not None
    assert item is not None

    name = next(k for k in item.assets.keys() if item.assets[k] == asset)
    item.delete_asset(name)

    assert name not in item.assets
    assert not os.path.exists(href)


def test_delete_asset_relative_no_self_link_fails(tmp_asset: pystac.Asset) -> None:
    asset = tmp_asset
    href = asset.get_absolute_href()
    item = asset.owner

    assert href is not None
    assert item is not None
    assert isinstance(item, pystac.Item)

    item.set_self_href(None)

    name = next(k for k in item.assets.keys() if item.assets[k] == asset)
    with pytest.raises(ValueError, match="Cannot delete file") as e:
        item.delete_asset(name)

    assert asset.href in str(e.value)
    assert name in item.assets
    assert os.path.exists(href)


def test_resolve_collection_with_root(
    tmp_path: Path, item: Item, collection: Collection
) -> None:
    # Motivated by https://github.com/stac-utils/pystac-client/issues/548
    catalog = Catalog("root", "the description")
    item.set_root(catalog)

    collection_path = str(tmp_path / "collection.json")
    collection.save_object(
        include_self_link=False,
        dest_href=collection_path,
    )
    item.add_link(Link(rel="collection", target=collection_path))

    read_collection = item.get_collection()
    assert read_collection
    root = read_collection.get_root()
    assert root
    assert root.id == "root"


@pytest.mark.vcr()
def test_non_hierarchical_relative_link() -> None:
    root = pystac.Catalog("root", "root")
    a = pystac.Catalog("a", "a")
    b = pystac.Catalog("b", "b")

    root.add_child(a)
    root.add_child(b)
    a.add_link(pystac.Link("related", b))
    b.add_link(
        pystac.Link("item", TestCases.get_path("data-files/item/sample-item.json"))
    )

    root.catalog_type = pystac.catalog.CatalogType.SELF_CONTAINED
    root.normalize_hrefs("test_output")
    related_href = [link for link in a.links if link.rel == "related"][0].get_href()

    assert related_href is not None and not is_absolute_href(related_href)
    assert a.target_in_hierarchy(b)
    assert root.target_in_hierarchy(next(b.get_items()))
    assert root.target_in_hierarchy(root)


def test_pathlib() -> None:
    # This works, but breaks mypy until we fix
    # https://github.com/stac-utils/pystac/issues/1216
    Item.from_file(Path(TestCases.get_path("data-files/item/sample-item.json")))


def test_invalid_error_message(item: Item) -> None:
    item.extra_fields["collection"] = "can't have a collection"
    with pytest.raises(STACValidationError) as error:
        item.validate()
    assert "can't have a collection" in str(error.value)


def test_pickle_with_no_links(item: Item) -> None:
    roundtripped = pickle.loads(pickle.dumps(item))
    for attr in ["id", "geometry", "bbox", "datetime", "links"]:
        assert getattr(roundtripped, attr) == getattr(item, attr)


def test_pickle_with_hrefless_links(item: Item) -> None:
    root = pystac.Catalog("root", "root")
    a = pystac.Catalog("a", "a")

    item.add_link(pystac.Link("related", a))
    item.add_link(
        pystac.Link("item", TestCases.get_path("data-files/item/sample-item.json"))
    )
    item.set_root(root)

    roundtripped = pickle.loads(pickle.dumps(item))
    for original, new in zip(item.links, roundtripped.links):
        assert original.rel == new.rel
        assert original.media_type == new.media_type
        assert str(original.owner) == str(new.owner)
        assert str(original.target) == str(new.target)


def test_pickle_with_only_href_links(item: Item) -> None:
    item.add_link(
        pystac.Link("item", TestCases.get_path("data-files/item/sample-item.json"))
    )

    roundtripped = pickle.loads(pickle.dumps(item))
    for original, new in zip(item.links, roundtripped.links):
        assert original.rel == new.rel
        assert original.media_type == new.media_type
        assert str(original.owner) == str(new.owner)
        assert str(original.target) == str(new.target)


def test_copy_with_unresolveable_root(item: Item) -> None:
    item.add_link(
        pystac.Link(
            "root", "s3://naip-visualization/this-is-a-non-existent-catalog.json"
        )
    )
    copy.deepcopy(item)


def test_no_collection(item: Item) -> None:
    # https://github.com/stac-utils/stac-api-validator/issues/527
    assert item.collection is None


def test_migrate_by_default() -> None:
    with open(
        TestCases.get_path("data-files/projection/example-with-version-1.1.json")
    ) as f:
        data = json.load(f)
    item = pystac.Item.from_dict(data)  # default used to be migrate=False
    assert item.ext.proj.code == "EPSG:32614"


def test_clone_extra_fields(item: Item) -> None:
    item.extra_fields["foo"] = "bar"
    cloned = item.clone()
    assert cloned.extra_fields["foo"] == "bar"



"""Implements the :stac-ext:`Projection Extension <projection>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.serialization.identify import STACJSONDescription, STACVersionID

#: Generalized version of :class:`~pystac.Item`, :class:`~pystac.Asset`,
#: or :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", pystac.Item, pystac.Asset, pystac.ItemAssetDefinition)

SCHEMA_URI: str = "https://stac-extensions.github.io/projection/v2.0.0/schema.json"
SCHEMA_URIS: list[str] = [
    "https://stac-extensions.github.io/projection/v1.0.0/schema.json",
    "https://stac-extensions.github.io/projection/v1.1.0/schema.json",
    "https://stac-extensions.github.io/projection/v1.2.0/schema.json",
    SCHEMA_URI,
]
PREFIX: str = "proj:"

# Field names
CODE_PROP: str = PREFIX + "code"
EPSG_PROP: str = PREFIX + "epsg"
WKT2_PROP: str = PREFIX + "wkt2"
PROJJSON_PROP: str = PREFIX + "projjson"
GEOM_PROP: str = PREFIX + "geometry"
BBOX_PROP: str = PREFIX + "bbox"
CENTROID_PROP: str = PREFIX + "centroid"
SHAPE_PROP: str = PREFIX + "shape"
TRANSFORM_PROP: str = PREFIX + "transform"


class ProjectionExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` with properties from the :stac-ext:`Projection
    Extension <projection>`. This class is generic over the type of STAC Object to be
    extended (e.g. :class:`~pystac.Item`, :class:`~pystac.Collection`).

    To create a concrete instance of :class:`ProjectionExtension`, use the
    :meth:`ProjectionExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> proj_ext = ProjectionExtension.ext(item)
    """

    name: Literal["proj"] = "proj"

    def apply(
        self,
        *,
        epsg: int | None = None,
        code: str | None = None,
        wkt2: str | None = None,
        projjson: dict[str, Any] | None = None,
        geometry: dict[str, Any] | None = None,
        bbox: list[float] | None = None,
        centroid: dict[str, float] | None = None,
        shape: list[int] | None = None,
        transform: list[float] | None = None,
    ) -> None:
        """Applies Projection extension properties to the extended Item.

        Args:
            epsg : Code of the datasource. Example: 4326.  One of ``code`` and
                ``epsg`` must be provided.
            code : Code of the datasource. Example: "EPSG:4326". One of ``code`` and
                ``epsg`` must be provided.
            wkt2 : WKT2 string representing the Coordinate Reference
                System (CRS) that the ``geometry`` and ``bbox`` fields represent
            projjson : PROJJSON dict representing the
                Coordinate Reference System (CRS) that the ``geometry`` and ``bbox``
                fields represent
            geometry : GeoJSON Polygon dict that defines the footprint of
                this Item.
            bbox : Bounding box of the Item in the asset CRS in
                2 or 3 dimensions.
            centroid : A dict with members 'lat' and 'lon' that defines
                coordinates representing the centroid of the item in the asset data CRS.
                Coordinates are defined in latitude and longitude, even if the data
                coordinate system may not use lat/long.
            shape : Number of pixels in Y and X directions for the
                default grid.
            transform : The affine transformation coefficients for
                the default grid
        """
        if epsg is not None and code is not None:
            raise KeyError(
                "Only one of the options ``code`` and ``epsg`` should be specified."
            )
        elif epsg:
            self.epsg = epsg
        else:
            self.code = code

        self.wkt2 = wkt2
        self.projjson = projjson
        self.geometry = geometry
        self.bbox = bbox
        self.centroid = centroid
        self.shape = shape
        self.transform = transform

    @property
    def epsg(self) -> int | None:
        """Get or sets the EPSG code of the datasource.

        A Coordinate Reference System (CRS) is the data reference system (sometimes
        called a 'projection') used by the asset data, and can usually be referenced
        using an `EPSG code <https://epsg.io/>`_.
        If the asset data does not have a CRS, such as in the case of non-rectified
        imagery with Ground Control Points, ``epsg`` should be set to ``None``.
        It should also be set to ``None`` if a CRS exists, but for which there is no
        valid EPSG code.
        """
        if self.code is not None and self.code.startswith("EPSG:"):
            return int(self.code.replace("EPSG:", ""))
        elif epsg := self._get_property(
            EPSG_PROP, int
        ):  # In case the dictionary was not migrated
            return epsg
        return None

    @epsg.setter
    def epsg(self, v: int | None) -> None:
        if v is None:
            self.code = None
        else:
            self.code = f"EPSG:{v}"

    @property
    def code(self) -> str | None:
        """Get or set the code of the datasource.

        Added in version 2.0.0 of this extension replacing "proj:epsg".

        Projection codes are identified by a string. The `proj <https://proj.org/>`_
        library defines projections using "authority:code", e.g., "EPSG:4326" or
        "IAU_2015:30100". Different projection authorities may define different
        string formats.
        """
        return self._get_property(CODE_PROP, str)

    @code.setter
    def code(self, v: str | None) -> None:
        self._set_property(CODE_PROP, v, pop_if_none=False)

    @property
    def wkt2(self) -> str | None:
        """Get or sets the WKT2 string representing the Coordinate Reference System
        (CRS) that the ``proj:geometry`` and ``proj:bbox`` fields represent

        This value is a
        `WKT2 string <https://docs.opengeospatial.org/is/12-063r5/12-063r5.html>`_.
        If the data does not have a CRS, such as in the case of non-rectified imagery
        with Ground Control Points, ``wkt2`` should be set to ``None``. It should also
        be set to ``None`` if a CRS exists, but for which a WKT2 string does not exist.
        """
        return self._get_property(WKT2_PROP, str)

    @wkt2.setter
    def wkt2(self, v: str | None) -> None:
        self._set_property(WKT2_PROP, v)

    @property
    def projjson(self) -> dict[str, Any] | None:
        """Get or sets the PROJJSON string representing the Coordinate Reference System
        (CRS) that the ``proj:geometry`` and ``proj:bbox`` fields represent

        This value is a
        `PROJJSON object <https://proj.org/specifications/projjson.html>`_.
        If the data does not have a CRS, such as in the case of non-rectified imagery
        with Ground Control Points, ``projjson`` should be set to ``None``. It should
        also be set to ``None`` if a CRS exists, but for which a PROJJSON string does
        not exist.

        The schema for this object can be found
        `here <https://proj.org/schemas/v0.2/projjson.schema.json>`_.
        """
        return self._get_property(PROJJSON_PROP, dict[str, Any])

    @projjson.setter
    def projjson(self, v: dict[str, Any] | None) -> None:
        self._set_property(PROJJSON_PROP, v)

    @property
    def crs_string(self) -> str | None:
        """Returns the coordinate reference system (CRS) string for the extension.

        This string can be used to feed, e.g., ``rasterio.crs.CRS.from_string``.
        The string is determined by the following heuristic:

        1. If a code is set, return the code string, else
        2. If wkt2 is set, return the WKT string, else,
        3. If projjson is set, return the projjson as a string, else,
        4. Return None
        """
        if self.code:
            return self.code
        elif self.epsg:
            return f"EPSG:{self.epsg}"
        elif self.wkt2:
            return self.wkt2
        elif self.projjson:
            import json

            return json.dumps(self.projjson)
        else:
            return None

    @property
    def geometry(self) -> dict[str, Any] | None:
        """Get or sets a Polygon GeoJSON dict representing the footprint of this item.

        This dict should be formatted according the Polygon object format specified in
        `RFC 7946, sections 3.1.6 <https://tools.ietf.org/html/rfc7946>`_,
        except not necessarily in EPSG:4326 as required by RFC7946. Specified based on
        the ``code``, ``projjson`` or ``wkt2`` fields (not necessarily EPSG:4326).
        Ideally, this will be represented by a Polygon with five coordinates, as the
        item in the asset data CRS should be a square aligned to the original CRS grid.
        """
        return self._get_property(GEOM_PROP, dict[str, Any])

    @geometry.setter
    def geometry(self, v: dict[str, Any] | None) -> None:
        self._set_property(GEOM_PROP, v)

    @property
    def bbox(self) -> list[float] | None:
        """Get or sets the bounding box of the assets represented by this item in the
        asset data CRS.

        Specified as 4 or 6 coordinates based on the CRS defined in the ``code``,
        ``projjson`` or ``wkt2`` properties. First two numbers are coordinates of the
        lower left corner, followed by coordinates of upper right corner, e.g.,
        ``[west, south, east, north]``, ``[xmin, ymin, xmax, ymax]``,
        ``[left, down, right, up]``, or ``[west, south, lowest, east, north,
        highest]``. The length of the array must be 2*n where n is the number of
        dimensions.
        """
        return self._get_property(BBOX_PROP, list[float])

    @bbox.setter
    def bbox(self, v: list[float] | None) -> None:
        self._set_property(BBOX_PROP, v)

    @property
    def centroid(self) -> dict[str, float] | None:
        """Get or sets coordinates representing the centroid of the item in the asset
        data CRS.

        Coordinates are defined in latitude and longitude, even if the data coordinate
        system does not use lat/long.

        Example::

            item.ext.proj.centroid = { 'lat': 0.0, 'lon': 0.0 }
        """
        return self._get_property(CENTROID_PROP, dict[str, float])

    @centroid.setter
    def centroid(self, v: dict[str, float] | None) -> None:
        self._set_property(CENTROID_PROP, v)

    @property
    def shape(self) -> list[int] | None:
        """Get or sets the number of pixels in Y and X directions for the default grid.

        The shape is an array of integers that represents the number of pixels in the
        most common pixel grid used by the item's assets. The number of pixels should
        be specified in Y, X order. If the shape is defined in an item's properties it
        is used as the default shape for all assets that don't have an overriding shape.
        """
        return self._get_property(SHAPE_PROP, list[int])

    @shape.setter
    def shape(self, v: list[int] | None) -> None:
        self._set_property(SHAPE_PROP, v)

    @property
    def transform(self) -> list[float] | None:
        """Get or sets the affine transformation coefficients for the default grid.

        The transform is a linear mapping from pixel coordinate space (Pixel, Line) to
        projection coordinate space (Xp, Yp). It is a 3x3 matrix stored as a flat array
        of 9 elements in row major order. Since the last row is always 0,0,1 it can be
        omitted, in which case only 6 elements are recorded. This mapping can be
        obtained from GDAL `GetGeoTransform <https://gdal.org/api/gdaldataset_cpp.html\
#_CPPv4N11GDALDataset15GetGeoTransformEPd>`_
        or the Rasterio `Transform <https://rasterio.readthedocs.io/en/stable/api\
/rasterio.io.html#rasterio.io.BufferedDatasetWriter.transform>`_.
        """
        return self._get_property(TRANSFORM_PROP, list[float])

    @transform.setter
    def transform(self, v: list[float] | None) -> None:
        self._set_property(TRANSFORM_PROP, v)

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def get_schema_uris(cls) -> list[str]:
        import warnings

        warnings.warn(
            "get_schema_uris is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return SCHEMA_URIS

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> ProjectionExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Projection
        Extension <projection>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(ProjectionExtension[T], ItemProjectionExtension(obj))
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(ProjectionExtension[T], AssetProjectionExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(ProjectionExtension[T], ItemAssetsProjectionExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesProjectionExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesProjectionExtension(obj)


class ItemProjectionExtension(ProjectionExtension[pystac.Item]):
    """A concrete implementation of :class:`ProjectionExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include properties
    defined in the :stac-ext:`Projection Extension <projection>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`ProjectionExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemProjectionExtension Item id={self.item.id}>"


class AssetProjectionExtension(ProjectionExtension[pystac.Asset]):
    """A concrete implementation of :class:`ProjectionExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include properties defined
    in the :stac-ext:`Projection Extension <projection>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`ProjectionExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetProjectionExtension Asset href={self.asset_href}>"


class ItemAssetsProjectionExtension(ProjectionExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class SummariesProjectionExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Projection Extension <projection>`.
    """

    @property
    def code(self) -> list[str] | None:
        """Get or sets the summary of :attr:`ProjectionExtension.code` values
        for this Collection.
        """
        return self.summaries.get_list(CODE_PROP)

    @code.setter
    def code(self, v: list[str] | None) -> None:
        self._set_summary(CODE_PROP, v)

    @property
    def epsg(self) -> list[int] | None:
        """Get the summary of :attr:`ProjectionExtension.epsg` values
        for this Collection.
        """
        if self.code is None:
            return None
        return [int(code.replace("EPSG:", "")) for code in self.code if "EPSG:" in code]

    @epsg.setter
    def epsg(self, v: list[int] | None) -> None:
        if v is None:
            self.code = None
        else:
            self.code = [f"EPSG:{epsg}" for epsg in v]


class ProjectionExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {
        "proj",
        "projection",
        *[uri for uri in SCHEMA_URIS if uri != SCHEMA_URI],
    }
    pre_2 = {
        "proj",
        "projection",
        "https://stac-extensions.github.io/projection/v1.0.0/schema.json",
        "https://stac-extensions.github.io/projection/v1.1.0/schema.json",
        "https://stac-extensions.github.io/projection/v1.2.0/schema.json",
    }
    stac_object_types = {pystac.STACObjectType.ITEM}

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        if not self.has_extension(obj):
            return

        import warnings

        # proj:epsg moved to proj:code
        if epsg := obj["properties"].pop("proj:epsg", None):
            if obj["properties"].get("proj:code", None) is None:
                obj["properties"]["proj:code"] = f"EPSG:{epsg}"
            elif not obj["properties"]["proj:code"] == f"EPSG:{epsg}":
                warnings.warn(
                    "Both proj:code and proj:epsg are specified and they have "
                    "conflicting values. This might lead to surprising behavior.",
                    UserWarning,
                )

        for key in ["assets", "item_assets"]:
            for asset in obj.get(key, {}).values():
                if epsg := asset.pop("proj:epsg", None):
                    if asset.get("proj:code", None) is None:
                        asset["proj:code"] = f"EPSG:{epsg}"
                    elif not asset["proj:code"] == f"EPSG:{epsg}":
                        warnings.warn(
                            "Both proj:code and proj:epsg are specified and they "
                            "have conflicting values. This might lead to surprising "
                            "behavior.",
                            UserWarning,
                        )

        super().migrate(obj, version, info)


PROJECTION_EXTENSION_HOOKS: ExtensionHooks = ProjectionExtensionHooks()



import json
import os
import shutil
import tempfile

from pystac import Collection, StacIO

from ._base import Bench
from ._util import get_data_path


class CollectionBench(Bench):
    def setup(self) -> None:
        self.temp_dir = tempfile.mkdtemp()

        self.stac_io = StacIO.default()

        self.collection_path = get_data_path("examples/1.0.0/collection.json")
        with open(self.collection_path) as src:
            self.collection_dict = json.load(src)
        self.collection = Collection.from_file(self.collection_path)

    def teardown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def time_collection_from_file(self) -> None:
        """Deserialize an Item from file"""
        _ = Collection.from_file(self.collection_path)

    def time_collection_from_dict(self) -> None:
        """Deserialize an Item from dictionary."""
        _ = Collection.from_dict(self.collection_dict)

    def time_collection_to_dict(self) -> None:
        """Serialize an Item to a dictionary."""
        self.collection.to_dict(include_self_link=True)

    def time_collection_save(self) -> None:
        """Serialize an Item to a JSON file."""
        self.collection.save_object(
            include_self_link=True,
            dest_href=os.path.join(self.temp_dir, "time_collection_save.json"),
            stac_io=self.stac_io,
        )



from datetime import datetime

from pystac import Item
from pystac.extensions.projection import ProjectionExtension

from .._base import Bench


class ProjectionBench(Bench):
    def setup(self) -> None:
        self.item = Item("an-id", None, None, datetime.now(), {})

    def time_add_projection_extension(self) -> None:
        _ = ProjectionExtension.ext(self.item, add_if_missing=True)
